import argparse
import os
import configparser
import tempfile
import sys
import subprocess
from pathlib import Path
import shutil
from enum import Enum
import time
from .consts import BASE_DIR_FOR_APP_ASSETS, APP_DAEMON_WORKSTAION_PATH


class SupervisorClientException(Exception):
    pass


class SupervisorClient:
    """
    A client for starting and stopping apps using supervisor.
    """

    def __init__(self, wait_time_seconds_for_app_start: int):
        self.supervisor_conf_loc = os.environ.get("SUPERVISOR_CONF_PATH")

        self.wait_time_seconds_for_app_start = wait_time_seconds_for_app_start
        if self.supervisor_conf_loc is None or not os.path.exists(
            self.supervisor_conf_loc
        ):
            raise SupervisorClientException(
                "This workstation does not support deploying apps! Please reach out to Outerbounds for support."
            )

        self.metaflow_envs_persistent_path = os.environ.get(
            "SUPERVISOR_PYTHON_ENVS_PATH"
        )
        if self.metaflow_envs_persistent_path is None:
            raise SupervisorClientException(
                "This workstation does not support deploying apps! Please reach out to Outerbounds for support."
            )

        # Check if supervisorctl is installed
        if not shutil.which("supervisorctl"):
            raise SupervisorClientException(
                "This workstation does not support deploying apps! Please reach out to Outerbounds for support."
            )

    def _stop_existing_app_at_port(self, app_port):
        supervisor_config = configparser.ConfigParser()
        supervisor_config.read(self.supervisor_conf_loc)

        for program in supervisor_config.sections():
            if "obp_app_port" in supervisor_config[program]:
                if supervisor_config[program]["obp_app_port"].strip() == str(app_port):
                    res = subprocess.run(
                        ["supervisorctl", "stop", program],
                        stdout=subprocess.DEVNULL,
                        stderr=subprocess.DEVNULL,
                    )

                    del supervisor_config[program]

        with tempfile.NamedTemporaryFile(
            "w", dir=os.path.dirname(self.supervisor_conf_loc), delete=False
        ) as f:
            supervisor_config.write(f)
            tmp_file = f.name

        os.rename(tmp_file, self.supervisor_conf_loc)

    def _create_supervisor_conf_entry(
        self, command, launch_directory, app_port, app_name
    ):
        entry = {
            "command": command,
            "directory": launch_directory,
            "autostart": "true",
            "autorestart": "true",
            "obp_app_port": app_port,  # Record the app port for internal reference. This is not used by supervisor.
        }

        supervisor_config = configparser.ConfigParser()
        supervisor_config.read(self.supervisor_conf_loc)

        supervisor_config[f"program:{app_name}"] = entry

        with tempfile.NamedTemporaryFile(
            "w", dir=os.path.dirname(self.supervisor_conf_loc), delete=False
        ) as f:
            supervisor_config.write(f)
            tmp_file = f.name

        os.rename(tmp_file, self.supervisor_conf_loc)

    def start_process_with_supervisord(
        self,
        app_name,
        app_port,
        user_provided_entrypoint,
        deploy_dir=None,
        app_dir=None,
    ):
        """
        Starts the app using supervisor.

        Args:
            app_name: The name of the app to start.
            app_port: The port to start the app on.
            user_provided_entrypoint: The entrypoint to start the app with.
            deploy_dir: The directory to copy the app to and deploy from.
            app_dir: The directory to copy the app from.
        """

        entrypoint = user_provided_entrypoint
        deploy_dir_for_port = os.path.join(BASE_DIR_FOR_APP_ASSETS, str(app_port))
        launch_directory = (
            BASE_DIR_FOR_APP_ASSETS
            if entrypoint is None
            else APP_DAEMON_WORKSTAION_PATH
        )

        # Stop any existing apps that are running on the same port.
        self._stop_existing_app_at_port(app_port)

        # This means the user has opted for either case 2 or case 3.
        # Cases 2 and 3 are handled the same way, the only thing that differs is self.app_dir.
        if user_provided_entrypoint is None:
            # Copy the app_dir to the deploy_dir.
            # This is also where all (if any) artifacts are written.
            recursive_copy(app_dir, deploy_dir)

            # Copy the entire deploy_dir to the port specific directory.
            # Clear out anything that was there before (maybe a different app's assets)
            if os.path.exists(deploy_dir_for_port):
                shutil.rmtree(deploy_dir_for_port)

            os.makedirs(deploy_dir_for_port)
            recursive_copy(deploy_dir, deploy_dir_for_port)

            # Apply default value
            # We launch the module from BASE_DIR_FOR_APP_ASSETS, so when we provide the -m flag we don't need (and can't use) the full path.
            # We just need to provide the port number (which is also the name of the folder where all app assets are stored)
            entrypoint = f"-m {str(app_port)}"

        # deploy_dir is meant to be temporary. No need to keep it around after everything has been copied over.
        shutil.rmtree(deploy_dir)

        # Metaflow by default generates the environment in /root/... (which is not persisted on workstations).
        # Since the environment is fully self contained, we can copy it to a persistent location.
        persistent_path_for_executable = (
            self._persist_metaflow_generated_python_environment()
        )

        # This is the command used by supervisord to launch the app.
        command = f"{persistent_path_for_executable} {entrypoint}"

        self._create_supervisor_conf_entry(
            command, launch_directory, app_port, app_name
        )

        # Execute supervisorctl reload
        # Capture the exit code
        exit_code = subprocess.run(
            ["supervisorctl", "reload"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode
        if exit_code != 0:
            raise SupervisorClientException(
                "Failed to start app! Contact Outerbounds for support."
            )

        print(
            f"Waiting for {self.wait_time_seconds_for_app_start} seconds for {app_name} to start..."
        )
        time.sleep(self.wait_time_seconds_for_app_start)

        self._raise_on_bad_status(app_name, command)

    def _get_launched_prcoess_status(self, app_name, debug_command):
        """
        Checks the status of the launched process. If the status is not RUNNING or STARTING, it raises an exception.
        Possible statuses: RUNNING, STARTING, STOPPED, BACKOFF, STOPPING, EXITED, FATAL, UNKNOWN

        Args:
            app_name: The name of the app to check the status of.
            debug_command: The command to run to debug the app.
        """
        status_cmd_output = subprocess.run(
            ["supervisorctl", "status", app_name],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ).stdout.decode("utf-8")

        status_cmd_output_parts = [
            x.strip() for x in status_cmd_output.split(" ") if x.strip()
        ]

        status_str = status_cmd_output_parts[1]

        if not status_str == "RUNNING" and not status_str == "STARTING":
            raise SupervisorClientException(
                f"Failed to start {app_name}! Try running {debug_command} manually to debug."
            )

    def _persist_metaflow_generated_python_environment(self):
        """
        Persists the metaflow generated python environment to a persistent location.
        The step already runs in the environment generated by Metaflow.
        """
        current_executable = sys.executable
        environment_path = Path(current_executable).parent.parent

        persistent_path_for_this_environment = os.path.join(
            self.metaflow_envs_persistent_path,
            environment_path.parent.name,
            environment_path.name,
        )

        final_executable_path = os.path.join(
            persistent_path_for_this_environment,
            Path(current_executable).parent.name,
            Path(current_executable).name,
        )

        if os.path.exists(final_executable_path):
            return final_executable_path

        os.makedirs(persistent_path_for_this_environment, exist_ok=True)

        recursive_copy(environment_path, persistent_path_for_this_environment)

        return final_executable_path


def recursive_copy(src, dst):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, dirs_exist_ok=True)
        else:
            shutil.copy2(s, d)



import json
import os
import sys
import random
from functools import wraps, partial
from typing import Dict, List, Any, Optional, Union

# IF this CLI is supposed to be reusable across Metaflow Too
# Then we need to ensure that the click IMPORT happens from metaflow
# so that any object class comparisons end up working as expected.
# since Metaflow lazy loads Click Modules, we need to ensure that the
# module's click Groups correlate to the same module otherwise Metaflow
# will not accept the cli as valid.
# But the BIGGEST Problem of adding a `metaflow._vendor` import is that
# It will run the remote-config check and that can raise a really dirty exception
# That will break the CLI.
# So we need to find a way to import click without having it try to check for remote-config
# or load the config from the environment.
# If we import click from metaflow over here then it might
# end up creating issues with click in general. So we need to figure a
# way to figure the right import of click dynamically. a neat way to handle that would be
# to have a function that can import the correct click based on the context in which stuff is being loaded.
from .click_importer import click
from .app_config import (
    AppConfig,
    AppConfigError,
    CODE_PACKAGE_PREFIX,
    CAPSULE_DEBUG,
    AuthType,
)
from .config import auto_cli_options, CoreConfig
from .perimeters import PerimeterExtractor
from .utils import (
    MultiStepSpinner,
)
from .code_package import CodePackager
from .capsule import (
    CapsuleDeployer,
    list_and_filter_capsules,
    CapsuleApi,
    DEPLOYMENT_READY_CONDITIONS,
    CapsuleApiException,
    CapsuleDeploymentException,
)
from .dependencies import bake_deployment_image
import shlex
import time
import uuid
from datetime import datetime


class KeyValueDictPair(click.ParamType):
    name = "KV-DICT-PAIR"  # type: ignore

    def convert(self, value, param, ctx):
        # Parse a string of the form KEY=VALUE into a dict {KEY: VALUE}
        if len(value.split("=", 1)) != 2:
            self.fail(
                f"Invalid format for {value}. Expected format: KEY=VALUE", param, ctx
            )

        key, _value = value.split("=", 1)
        try:
            return {"key": key, "value": json.loads(_value)}
        except json.JSONDecodeError:
            return {"key": key, "value": _value}
        except Exception as e:
            self.fail(f"Invalid value for {value}. Error: {e}", param, ctx)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "KV-PAIR"


class KeyValuePair(click.ParamType):
    name = "KV-PAIR"  # type: ignore

    def convert(self, value, param, ctx):
        # Parse a string of the form KEY=VALUE into a dict {KEY: VALUE}
        if len(value.split("=", 1)) != 2:
            self.fail(
                f"Invalid format for {value}. Expected format: KEY=VALUE", param, ctx
            )

        key, _value = value.split("=", 1)
        try:
            return {key: json.loads(_value)}
        except json.JSONDecodeError:
            return {key: _value}
        except Exception as e:
            self.fail(f"Invalid value for {value}. Error: {e}", param, ctx)

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "KV-PAIR"


class CommaSeparatedList(click.ParamType):
    name = "COMMA-SEPARATED-LIST"  # type: ignore

    def convert(self, value, param, ctx):
        return value.split(",")

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return "COMMA-SEPARATED-LIST"


KVPairType = KeyValuePair()  # used for --tag and --env
CommaSeparatedListType = CommaSeparatedList()  # used for --compute-pools
KVDictType = KeyValueDictPair()  # only Used for the list/delete commands for tags


class ColorTheme:
    TIMESTAMP = "magenta"
    LOADING_COLOR = "cyan"
    BAD_COLOR = "red"
    INFO_COLOR = "green"
    DEBUG_COLOR = "yellow"

    TL_HEADER_COLOR = "magenta"
    # Use a color that is readable in both light and dark mode.
    # "white" can be hard to see in light mode, so use "black" for rows.
    # Alternatively, "reset" (default terminal color) is safest.
    ROW_COLOR = "reset"

    INFO_KEY_COLOR = "green"
    INFO_VALUE_COLOR = "reset"


NativeList = list


def _logger(
    body="", system_msg=False, head="", bad=False, timestamp=True, nl=True, color=None
):
    if timestamp:
        if timestamp is True:
            dt = datetime.now()
        else:
            dt = timestamp
        tstamp = dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        click.secho(tstamp + " ", fg=ColorTheme.TIMESTAMP, nl=False)
    if head:
        click.secho(head, fg=ColorTheme.INFO_COLOR, nl=False)
    click.secho(
        body,
        bold=system_msg,
        fg=ColorTheme.BAD_COLOR if bad else color if color is not None else None,
        nl=nl,
    )


def _logger_styled(
    body="", system_msg=False, head="", bad=False, timestamp=True, nl=True, color=None
):
    message_parts = []

    if timestamp:
        if timestamp is True:
            dt = datetime.now()
        else:
            dt = timestamp
        tstamp = dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        message_parts.append(click.style(tstamp + " ", fg=ColorTheme.TIMESTAMP))

    if head:
        message_parts.append(click.style(head, fg=ColorTheme.INFO_COLOR))

    message_parts.append(
        click.style(
            body,
            bold=system_msg,
            fg=ColorTheme.BAD_COLOR if bad else color if color is not None else None,
        )
    )

    return "".join(message_parts)


def _spinner_logger(spinner, *msg, **kwargs):
    spinner.log(*[_logger_styled(x, timestamp=True, **kwargs) for x in msg])


class CliState(object):
    pass


def _pre_create_debug(
    app_config: AppConfig,
    capsule: CapsuleDeployer,
    state_dir: str,
    options: Dict[str, Any],
):
    if CAPSULE_DEBUG:
        os.makedirs(state_dir, exist_ok=True)
        debug_path = os.path.join(state_dir, f"debug_{time.time()}.json")
        with open(
            debug_path,
            "w",
        ) as f:
            f.write(
                json.dumps(
                    {
                        "app_state": app_config.dump_state(),  # This is the state of the app config after parsing the CLI options and right before the capsule deploy API is called
                        "capsule_input": capsule.create_input(),  # This is the input that is passed to the capsule deploy API
                        "deploy_response": capsule._capsule_deploy_response,  # This is the response from the capsule deploy API
                        "cli_options": options,  # These are the actual options passing down to the CLI
                    },
                    indent=2,
                    default=str,
                )
            )


def _post_create_debug(capsule: CapsuleDeployer, state_dir: str):
    if CAPSULE_DEBUG:
        debug_path = os.path.join(
            state_dir, f"debug_deploy_response_{time.time()}.json"
        )
        with open(debug_path, "w") as f:
            f.write(json.dumps(capsule._capsule_deploy_response, indent=2, default=str))


def _bake_image(app_config: AppConfig, cache_dir: str, logger):
    baking_status = bake_deployment_image(
        app_config=app_config,
        cache_file_path=os.path.join(cache_dir, "image_cache"),
        logger=logger,
    )
    app_config.set_state(
        "image",
        baking_status.resolved_image,
    )
    app_config.set_state("python_path", baking_status.python_path)
    logger("🐳 Using the docker image : %s" % app_config.get_state("image"))


def print_table(data, headers):
    """Print data in a formatted table."""

    if not data:
        return

    # Calculate column widths
    col_widths = [len(h) for h in headers]

    # Calculate actual widths based on data
    for row in data:
        for i, cell in enumerate(row):
            col_widths[i] = max(col_widths[i], len(str(cell)))

    # Print header
    header_row = " | ".join(
        [headers[i].ljust(col_widths[i]) for i in range(len(headers))]
    )
    click.secho("-" * len(header_row), fg=ColorTheme.TL_HEADER_COLOR)
    click.secho(header_row, fg=ColorTheme.TL_HEADER_COLOR, bold=True)
    click.secho("-" * len(header_row), fg=ColorTheme.TL_HEADER_COLOR)

    # Print data rows
    for row in data:
        formatted_row = " | ".join(
            [str(row[i]).ljust(col_widths[i]) for i in range(len(row))]
        )
        click.secho(formatted_row, fg=ColorTheme.ROW_COLOR, bold=True)
    click.secho("-" * len(header_row), fg=ColorTheme.TL_HEADER_COLOR)


@click.group()
def cli():
    """Outerbounds CLI tool."""
    pass


@cli.group(
    help="Commands related to Deploying/Running/Managing Apps on Outerbounds Platform."
)
@click.pass_context
def app(ctx):
    """App-related commands."""
    metaflow_set_context = getattr(ctx, "obj", None)
    ctx.obj = CliState()
    ctx.obj.trace_id = str(uuid.uuid4())
    ctx.obj.app_state_dir = os.path.join(os.curdir, ".ob_apps")
    profile = os.environ.get("METAFLOW_PROFILE", "")
    config_dir = os.path.expanduser(
        os.environ.get("METAFLOW_HOME", "~/.metaflowconfig")
    )
    perimeter, api_server = PerimeterExtractor.for_ob_cli(
        config_dir=config_dir, profile=profile
    )
    if perimeter is None or api_server is None:
        raise AppConfigError(
            "Perimeter not found in the environment, Found perimeter: %s, api_server: %s"
            % (perimeter, api_server)
        )
    ctx.obj.perimeter = perimeter
    ctx.obj.api_url = api_server
    os.makedirs(ctx.obj.app_state_dir, exist_ok=True)


def parse_cli_commands(cli_command_input):
    # There can be two modes:
    # 1. User passes command via `--` in the CLI
    # 2. User passes the `commands` key in the config.
    # This function parses the command for mode 1.
    base_commands = []
    if len(cli_command_input) > 0:
        if type(cli_command_input) == str:
            base_commands.append(cli_command_input)
        else:
            base_commands.append(shlex.join(cli_command_input))

    return base_commands


def deployment_instance_options(func):
    # These parameters influence how the CLI behaves for each instance of a launched deployment.
    @click.option(
        "--readiness-condition",
        type=click.Choice(DEPLOYMENT_READY_CONDITIONS.enums()),
        help=DEPLOYMENT_READY_CONDITIONS.__doc__,
        default=DEPLOYMENT_READY_CONDITIONS.ATLEAST_ONE_RUNNING,
    )
    @click.option(
        "--status-file",
        type=str,
        help="The path to the file where the final status of the deployment will be written.",
        default=None,
    )
    @click.option(
        "--readiness-wait-time",
        type=int,
        help="The time (in seconds) to monitor the deployment for readiness after the readiness condition is met.",
        default=4,
    )
    @click.option(
        "--deployment-timeout",
        "max_wait_time",
        type=int,
        help="The maximum time (in seconds) to wait for the deployment to reach readiness before timing out.",
        default=600,
    )
    @click.option(
        "--no-loader",
        is_flag=True,
        help="Do not use the loading spinner for the deployment.",
        default=False,
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def _package_necessary_things(app_config: AppConfig, logger):
    # Packaging has a few things to be thought through:
    #   1. if `entrypoint_path` exists then should we package the directory
    #      where the entrypoint lives. For example : if the user calls
    #      `outerbounds app deploy foo/bar.py`  should we package `foo` dir
    #      or should we package the cwd from which foo/bar.py is being called.
    #   2. if the src path is used with the config file then how should we view
    #      that path ?
    #   3. It becomes interesting when users call the deployment with config files
    #      where there is a `src_path` and then is the src_path relative to the config file
    #      or is it relative to where the caller command is sitting. Ideally it should work
    #      like Kustomizations where its relative to where the yaml file sits for simplicity
    #      of understanding relationships between config files. Ideally users can pass the src_path
    #      from the command line and that will alleviate any need to package any other directories for
    #

    package_dir = app_config.get_state("packaging_directory")
    if package_dir is None:
        app_config.set_state("code_package_url", None)
        app_config.set_state("code_package_key", None)
        return
    from metaflow.metaflow_config import DEFAULT_DATASTORE

    package = app_config.get_state("package") or {}
    suffixes = package.get("suffixes", None)

    packager = CodePackager(
        datastore_type=DEFAULT_DATASTORE, code_package_prefix=CODE_PACKAGE_PREFIX
    )
    package_url, package_key = packager.store(
        paths_to_include=[package_dir], file_suffixes=suffixes
    )
    app_config.set_state("code_package_url", package_url)
    app_config.set_state("code_package_key", package_key)
    logger("💾 Code package saved to : %s" % app_config.get_state("code_package_url"))


@app.command(help="Deploy an app to the Outerbounds Platform.")
@click.option(
    "--config-file",
    type=str,
    help="The config file to use for the App (YAML or JSON)",
    default=None,
)
@deployment_instance_options
@auto_cli_options()
@click.pass_context
@click.argument("command", nargs=-1, type=click.UNPROCESSED, required=False)
def deploy(
    ctx,
    command,
    readiness_condition=None,
    max_wait_time=None,
    readiness_wait_time=None,
    status_file=None,
    no_loader=False,
    **options,
):
    """Deploy an app to the Outerbounds Platform."""
    from functools import partial

    if not ctx.obj.perimeter:
        raise AppConfigError("OB_CURRENT_PERIMETER is not set")
    _current_instance_debug_dir = None
    logger = partial(_logger, timestamp=True)

    base_commands = parse_cli_commands(command)
    options["commands"] = base_commands
    try:
        # Create configuration
        if options["config_file"]:
            # Load from file
            app_config = AppConfig.from_file(options["config_file"])

            # Update with any CLI options using the unified method
            app_config.update_from_cli_options(options)
        else:
            # Create from CLI options
            app_config = AppConfig.from_cli(options)

        # Validate the configuration
        app_config.commit()
        logger(
            f"🚀 Deploying {app_config.get('name')} to the Outerbounds platform...",
            color=ColorTheme.INFO_COLOR,
            system_msg=True,
        )

        packaging_directory = None
        package_src_path = app_config.get("package", {}).get("src_path", None)
        if package_src_path:
            if os.path.isfile(package_src_path):
                raise AppConfigError("src_path must be a directory, not a file")
            elif os.path.isdir(package_src_path):
                packaging_directory = os.path.abspath(package_src_path)
            else:
                raise AppConfigError(f"src_path '{package_src_path}' does not exist")
        else:
            # If src_path is None then we assume then we can assume for the moment
            # that we can package the current working directory.
            packaging_directory = os.getcwd()

        app_config.set_state("packaging_directory", packaging_directory)
        logger(
            "📦 Packaging directory : %s" % app_config.get_state("packaging_directory"),
        )

        if app_config.get("no_deps", False):
            # Setting this in the state will make it skip the fast-bakery step
            # of building an image.
            app_config.set_state("skip_dependencies", True)
        else:
            # Check if the user has set the dependencies in the app config
            dependencies = app_config.get("dependencies", {})

            if all(
                [
                    dependencies.get("from_pyproject_toml", None) is None,
                    dependencies.get("from_requirements_file", None) is None,
                    dependencies.get("pypi", None) is None,
                    dependencies.get("conda", None) is None,
                ]
            ):
                python_version = dependencies.get(
                    "python"
                )  # python gets a default value so it's always set.
                # The user has not set any dependencies, so we can sniff the packaging directory
                # for a dependencies file.
                requirements_file = os.path.join(
                    packaging_directory, "requirements.txt"
                )
                pyproject_toml = os.path.join(packaging_directory, "pyproject.toml")
                if os.path.exists(pyproject_toml):
                    app_config.set_state(
                        "dependencies",
                        {
                            "from_pyproject_toml": pyproject_toml,
                            "python": python_version,
                        },
                    )
                    logger(
                        "📦 Using dependencies from pyproject.toml: %s" % pyproject_toml
                    )
                elif os.path.exists(requirements_file):
                    app_config.set_state(
                        "dependencies",
                        {
                            "from_requirements_file": requirements_file,
                            "python": python_version,
                        },
                    )
                    logger(
                        "📦 Using dependencies from requirements.txt: %s"
                        % requirements_file
                    )

        # Print the configuration
        # 1. validate that the secrets for the app exist
        # 2. TODO: validate that the compute pool specified in the app exists.
        # 3. Building Docker image if necessary (based on parameters)
        #   - We will bake images with fastbakery and pass it to the deploy command
        cache_dir = os.path.join(
            ctx.obj.app_state_dir, app_config.get("name", "default")
        )

        def _non_spinner_logger(*msg, **kwargs):
            for m in msg:
                logger(m, **kwargs)

        image_spinner = None
        img_logger = _non_spinner_logger
        if not no_loader:
            image_spinner = MultiStepSpinner(
                text=lambda: _logger_styled(
                    "🍞 Baking Docker Image",
                    timestamp=True,
                ),
                color=ColorTheme.LOADING_COLOR,
            )
            img_logger = partial(_spinner_logger, image_spinner)
            image_spinner.start()

        _bake_image(app_config, cache_dir, img_logger)
        if image_spinner:
            image_spinner.stop()

        # TODO: Handle the case where packaging_directory is None
        # This would involve:
        # 1. Packaging the code:
        #   - We need to package the code and throw the tarball to some object store
        _package_necessary_things(app_config, logger)

        app_config.set_state("perimeter", ctx.obj.perimeter)

        capsule_spinner = None
        capsule_logger = _non_spinner_logger

        if CAPSULE_DEBUG:
            _current_instance_debug_dir = os.path.join(
                cache_dir, f"debug_deployment_instance_{time.time()}"
            )
            os.makedirs(_current_instance_debug_dir, exist_ok=True)
        # 2. Convert to the IR that the backend accepts
        capsule = CapsuleDeployer(
            app_config,
            ctx.obj.api_url,
            debug_dir=_current_instance_debug_dir,
            success_terminal_state_condition=readiness_condition,
            create_timeout=max_wait_time,
            readiness_wait_time=readiness_wait_time,
            logger_fn=capsule_logger,
        )
        if not no_loader:
            capsule_spinner = MultiStepSpinner(
                text=lambda: _logger_styled(
                    "💊 Waiting for %s %s to be ready to serve traffic"
                    % (capsule.capsule_type.lower(), capsule.identifier),
                    timestamp=True,
                ),
                color=ColorTheme.LOADING_COLOR,
            )
            capsule_logger = partial(_spinner_logger, capsule_spinner)
            capsule_spinner.start()

        currently_present_capsules = list_and_filter_capsules(
            capsule.capsule_api,
            None,
            None,
            capsule.name,
            None,
            None,
            None,
        )

        force_upgrade = app_config.get_state("force_upgrade", False)

        _pre_create_debug(
            app_config,
            capsule,
            _current_instance_debug_dir,
            options,
        )

        if len(currently_present_capsules) > 0:
            # Only update the capsule if there is no upgrade in progress
            # Only update a "already updating" capsule if the `--force-upgrade` flag is provided.
            _curr_cap = currently_present_capsules[0]
            this_capsule_is_being_updated = _curr_cap.get("status", {}).get(
                "updateInProgress", False
            )

            if this_capsule_is_being_updated and not force_upgrade:
                _upgrader = _curr_cap.get("metadata", {}).get("lastModifiedBy", None)
                message = f"{capsule.capsule_type} is currently being upgraded"
                if _upgrader:
                    message = (
                        f"{capsule.capsule_type} is currently being upgraded. Upgrade was launched by {_upgrader}. "
                        "If you wish to force upgrade, you can do so by providing the `--force-upgrade` flag."
                    )
                raise AppConfigError(message)
            capsule_logger(
                f"🚀 {'Upgrading' if not force_upgrade else 'Force upgrading'} {capsule.capsule_type.lower()} `{capsule.name}`....",
                color=ColorTheme.INFO_COLOR,
                system_msg=True,
            )
        else:
            capsule_logger(
                f"🚀 Deploying {capsule.capsule_type.lower()} to the platform....",
                color=ColorTheme.INFO_COLOR,
                system_msg=True,
            )
        # 3. Throw the job into the platform and report deployment status
        capsule.create()
        _post_create_debug(capsule, _current_instance_debug_dir)

        # We only get the `capsule_response` if the deployment is has reached
        # a successful terminal state.
        final_status = capsule.wait_for_terminal_state()
        if capsule_spinner:
            capsule_spinner.stop()

        logger(
            f"💊 {capsule.capsule_type} {app_config.config['name']} ({capsule.identifier}) deployed! {capsule.capsule_type} available on the URL: {capsule.url}",
            color=ColorTheme.INFO_COLOR,
            system_msg=True,
        )

        if CAPSULE_DEBUG:
            logger(
                f"[debug] 💊 {capsule.capsule_type} {app_config.config['name']} ({capsule.identifier}) deployment status [on completion]: {final_status}",
                color=ColorTheme.DEBUG_COLOR,
            )
            logger(
                f"[debug] 💊 {capsule.capsule_type} {app_config.config['name']} ({capsule.identifier}) debug info saved to `{_current_instance_debug_dir}`",
                color=ColorTheme.DEBUG_COLOR,
            )
            final_status["debug_dir"] = _current_instance_debug_dir

        if status_file:
            # Create the file if it doesn't exist
            with open(status_file, "w") as f:
                f.write(json.dumps(final_status, indent=4))
            logger(
                f"📝 {capsule.capsule_type} {app_config.config['name']} ({capsule.identifier}) deployment status written to {status_file}",
                color=ColorTheme.INFO_COLOR,
                system_msg=True,
            )

    except Exception as e:
        message = getattr(e, "message", str(e))
        logger(
            f"Deployment failed: [{e.__class__.__name__}]: {message}",
            bad=True,
            system_msg=True,
        )
        if CAPSULE_DEBUG:
            if _current_instance_debug_dir is not None:
                logger(
                    f"[debug] 💊 debug info saved to `{_current_instance_debug_dir}`",
                    color=ColorTheme.DEBUG_COLOR,
                )
            raise e
        exit(1)


def _parse_capsule_table(filtered_capsules):
    headers = ["Name", "ID", "Ready", "App Type", "Port", "Tags", "URL"]
    table_data = []

    for capsule in filtered_capsules:
        spec = capsule.get("spec", {})
        status = capsule.get("status", {}) or {}
        cap_id = capsule.get("id")
        display_name = spec.get("displayName", "")
        ready = str(status.get("readyToServeTraffic", False))
        auth_type = spec.get("authConfig", {}).get("authType", "")
        port = str(spec.get("port", ""))
        tags_str = ", ".join(
            [f"{tag['key']}={tag['value']}" for tag in spec.get("tags", [])]
        )
        access_info = status.get("accessInfo", {}) or {}
        url = access_info.get("outOfClusterURL", None)

        table_data.append(
            [
                display_name,
                cap_id,
                ready,
                auth_type,
                port,
                tags_str,
                f"https://{url}" if url else "URL not available",
            ]
        )
    return headers, table_data


@app.command(help="List apps in the Outerbounds Platform.")
@click.option("--project", type=str, help="Filter apps by project")
@click.option("--branch", type=str, help="Filter apps by branch")
@click.option("--name", type=str, help="Filter apps by name")
@click.option(
    "--tag",
    "tags",
    type=KVDictType,
    help="Filter apps by tag. Format KEY=VALUE. Example --tag foo=bar --tag x=y. If multiple tags are provided, the app must match all of them.",
    multiple=True,
)
@click.option(
    "--format",
    type=click.Choice(["json", "text"]),
    help="Format the output",
    default="text",
)
@click.option(
    "--auth-type", type=click.Choice(AuthType.enums()), help="Filter apps by Auth type"
)
@click.pass_context
def list(ctx, project, branch, name, tags, format, auth_type):
    """List apps in the Outerbounds Platform."""
    capsule_api = CapsuleApi(
        ctx.obj.api_url,
        ctx.obj.perimeter,
    )
    filtered_capsules = list_and_filter_capsules(
        capsule_api, project, branch, name, tags, auth_type, None
    )
    if format == "json":
        click.echo(json.dumps(filtered_capsules, indent=4))
    else:
        headers, table_data = _parse_capsule_table(filtered_capsules)
        print_table(table_data, headers)


@app.command(help="Delete an app/apps from the Outerbounds Platform.")
@click.option("--name", type=str, help="Filter app to delete by name")
@click.option("--id", "cap_id", type=str, help="Filter app to delete by id")
@click.option("--project", type=str, help="Filter apps to delete by project")
@click.option("--branch", type=str, help="Filter apps to delete by branch")
@click.option(
    "--tag",
    "tags",
    multiple=True,
    type=KVDictType,
    help="Filter apps to delete by tag. Format KEY=VALUE. Example --tag foo=bar --tag x=y. If multiple tags are provided, the app must match all of them.",
)
@click.option("--auto-approve", is_flag=True, help="Do not prompt for confirmation")
@click.pass_context
def delete(ctx, name, cap_id, project, branch, tags, auto_approve):

    """Delete an app/apps from the Outerbounds Platform."""
    # At least one of the args need to be provided
    if not any(
        [
            name is not None,
            cap_id is not None,
            project is not None,
            branch is not None,
            len(tags) != 0,
        ]
    ):
        raise AppConfigError(
            "At least one of the options need to be provided. You can use --name, --id, --project, --branch, --tag"
        )

    capsule_api = CapsuleApi(ctx.obj.api_url, ctx.obj.perimeter)
    filtered_capsules = list_and_filter_capsules(
        capsule_api, project, branch, name, tags, None, cap_id
    )

    headers, table_data = _parse_capsule_table(filtered_capsules)
    click.secho("The following apps will be deleted:", fg="red", bold=True)
    print_table(table_data, headers)

    # Confirm the deletion
    if not auto_approve:
        confirm = click.prompt(
            click.style(
                "💊 Are you sure you want to delete these apps?", fg="red", bold=True
            ),
            default="no",
            type=click.Choice(["yes", "no"]),
        )
        if confirm == "no":
            exit(1)

    def item_show_func(x):
        if not x:
            return None
        name = x.get("spec", {}).get("displayName", "")
        id = x.get("id", "")
        return click.style(
            "💊 deleting %s [%s]" % (name, id),
            fg=ColorTheme.BAD_COLOR,
            bold=True,
        )

    with click.progressbar(
        filtered_capsules,
        label=click.style("💊 Deleting apps...", fg=ColorTheme.BAD_COLOR, bold=True),
        fill_char=click.style("█", fg=ColorTheme.BAD_COLOR, bold=True),
        empty_char=click.style("░", fg=ColorTheme.BAD_COLOR, bold=True),
        item_show_func=item_show_func,
    ) as bar:
        for capsule in bar:
            capsule_api.delete(capsule.get("id"))
            time.sleep(0.5 + random.random() * 2)  # delay to avoid rate limiting


@app.command(
    help="Get detailed information about an app from the Outerbounds Platform."
)
@click.option("--name", type=str, help="Get info for app by name")
@click.option("--id", "cap_id", type=str, help="Get info for app by id")
@click.option(
    "--format",
    type=click.Choice(["json", "text"]),
    help="Format the output",
    default="text",
)
@click.pass_context
def info(ctx, name, cap_id, format):
    """Get detailed information about an app from the Outerbounds Platform."""
    # Require either name or id
    if not any([name is not None, cap_id is not None]):
        raise AppConfigError(
            "Either --name or --id must be provided to get app information."
        )

    # Ensure only one is provided
    if name is not None and cap_id is not None:
        raise AppConfigError("Please provide either --name or --id, not both.")

    capsule_api = CapsuleApi(
        ctx.obj.api_url,
        ctx.obj.perimeter,
    )

    # First, find the capsule using list_and_filter_capsules
    filtered_capsules = list_and_filter_capsules(
        capsule_api, None, None, name, None, None, cap_id
    )

    if len(filtered_capsules) == 0:
        identifier = name if name else cap_id
        identifier_type = "name" if name else "id"
        raise AppConfigError(f"No app found with {identifier_type}: {identifier}")

    if len(filtered_capsules) > 1:
        raise AppConfigError(
            f"Multiple apps found with name: {name}. Please use --id to specify exactly which app you want info for."
        )

    # Get the capsule info
    capsule = filtered_capsules[0]
    capsule_id = capsule.get("id")

    # Get detailed capsule info and workers
    try:
        detailed_capsule_info = capsule_api.get(capsule_id)
        workers_info = capsule_api.get_workers(capsule_id)

        if format == "json":
            # Output in JSON format for piping to jq
            info_data = {"capsule": detailed_capsule_info, "workers": workers_info}
            click.echo(json.dumps(info_data, indent=4))
        else:
            # Output in text format
            _display_capsule_info_text(detailed_capsule_info, workers_info)

    except Exception as e:
        raise AppConfigError(f"Error retrieving information for app {capsule_id}: {e}")


def _display_capsule_info_text(capsule_info, workers_info):
    """Display capsule information in a human-readable text format."""
    spec = capsule_info.get("spec", {})
    status = capsule_info.get("status", {}) or {}
    metadata = capsule_info.get("metadata", {}) or {}

    info_color = ColorTheme.INFO_COLOR
    tl_color = ColorTheme.TL_HEADER_COLOR

    def _key_style(key: str, value: str):
        return "%s: %s" % (
            click.style(
                key,
                fg=ColorTheme.INFO_KEY_COLOR,
            ),
            click.style(str(value), fg=ColorTheme.INFO_VALUE_COLOR, bold=True),
        )

    # Basic Info
    click.secho("=== App Information ===", fg=tl_color, bold=True)
    click.secho(_key_style("Name", spec.get("displayName", "N/A")), fg=info_color)
    click.secho(_key_style("ID", capsule_info.get("id", "N/A")), fg=info_color)
    click.secho(
        _key_style("Version", capsule_info.get("version", "N/A")), fg=info_color
    )
    click.secho(
        _key_style(
            "Ready to Serve Traffic", str(status.get("readyToServeTraffic", False))
        ),
        fg=info_color,
    )
    click.secho(
        _key_style("Update In Progress", str(status.get("updateInProgress", False))),
        fg=info_color,
    )
    click.secho(
        _key_style(
            "Currently Served Version", str(status.get("currentlyServedVersion", "N/A"))
        ),
        fg=info_color,
    )

    # URLs
    access_info = status.get("accessInfo", {}) or {}
    out_cluster_url = access_info.get("outOfClusterURL")
    in_cluster_url = access_info.get("inClusterURL")

    if out_cluster_url:
        click.secho(
            _key_style("External URL", f"https://{out_cluster_url}"), fg=info_color
        )
    if in_cluster_url:
        click.secho(
            _key_style("Internal URL", f"https://{in_cluster_url}"), fg=info_color
        )

    # Resource Configuration
    click.secho("\n=== Resource Configuration ===", fg=tl_color, bold=True)
    resource_config = spec.get("resourceConfig", {})
    click.secho(_key_style("CPU", resource_config.get("cpu", "N/A")), fg=info_color)
    click.secho(
        _key_style("Memory", resource_config.get("memory", "N/A")), fg=info_color
    )
    click.secho(
        _key_style("Ephemeral Storage", resource_config.get("ephemeralStorage", "N/A")),
        fg=info_color,
    )
    if resource_config.get("gpu"):
        click.secho(_key_style("GPU", resource_config.get("gpu")), fg=info_color)

    # Autoscaling
    click.secho("\n=== Autoscaling Configuration ===", fg=tl_color, bold=True)
    autoscaling_config = spec.get("autoscalingConfig", {})
    click.secho(
        _key_style("Min Replicas", str(autoscaling_config.get("minReplicas", "N/A"))),
        fg=info_color,
    )
    click.secho(
        _key_style("Max Replicas", str(autoscaling_config.get("maxReplicas", "N/A"))),
        fg=info_color,
    )
    click.secho(
        _key_style("Available Replicas", str(status.get("availableReplicas", "N/A"))),
        fg=info_color,
    )

    # Auth Configuration
    click.secho("\n=== Authentication Configuration ===", fg=tl_color, bold=True)
    auth_config = spec.get("authConfig", {})
    click.secho(
        _key_style("Auth Type", auth_config.get("authType", "N/A")), fg=info_color
    )
    click.secho(
        _key_style("Public Access", str(auth_config.get("publicToDeployment", "N/A"))),
        fg=info_color,
    )

    # Tags
    tags = spec.get("tags", [])
    if tags:
        click.secho("\n=== Tags ===", fg=tl_color, bold=True)
        for tag in tags:
            click.secho(
                _key_style(str(tag.get("key", "N/A")), str(tag.get("value", "N/A"))),
                fg=info_color,
            )

    # Metadata
    click.secho("\n=== Metadata ===", fg=tl_color, bold=True)
    click.secho(
        _key_style("Created At", metadata.get("createdAt", "N/A")), fg=info_color
    )
    click.secho(
        _key_style("Last Modified At", metadata.get("lastModifiedAt", "N/A")),
        fg=info_color,
    )
    click.secho(
        _key_style("Last Modified By", metadata.get("lastModifiedBy", "N/A")),
        fg=info_color,
    )

    # Workers Information
    click.secho("\n=== Workers Information ===", fg=tl_color, bold=True)
    if not workers_info:
        click.secho("No workers found", fg=info_color)
    else:
        click.secho(_key_style("Total Workers", str(len(workers_info))), fg=tl_color)

        # Create a table for workers
        workers_headers = [
            "Worker ID",
            "Phase",
            "Version",
            "Activity",
            "Activity Data Available",
        ]
        workers_table_data = []

        for worker in workers_info:
            worker_id = worker.get("workerId", "N/A")
            phase = worker.get("phase", "N/A")
            version = worker.get("version", "N/A")
            activity = str(worker.get("activity", "N/A"))
            activity_data_available = str(worker.get("activityDataAvailable", False))

            workers_table_data.append(
                [
                    worker_id[:20] + "..." if len(worker_id) > 23 else worker_id,
                    phase,
                    version[:10] + "..." if len(version) > 13 else version,
                    activity,
                    activity_data_available,
                ]
            )

        print_table(workers_table_data, workers_headers)


@app.command(help="Get logs for an app worker from the Outerbounds Platform.")
@click.option("--name", type=str, help="Get logs for app by name")
@click.option("--id", "cap_id", type=str, help="Get logs for app by id")
@click.option("--worker-id", type=str, help="Get logs for specific worker")
@click.option("--file", type=str, help="Save logs to file")
@click.option(
    "--previous",
    is_flag=True,
    help="Get logs from previous container instance",
    default=False,
)
@click.pass_context
def logs(ctx, name, cap_id, worker_id, file, previous):
    """Get logs for an app worker from the Outerbounds Platform."""
    # Require either name or id
    if not any([name is not None, cap_id is not None]):
        raise AppConfigError("Either --name or --id must be provided to get app logs.")

    # Ensure only one is provided
    if name is not None and cap_id is not None:
        raise AppConfigError("Please provide either --name or --id, not both.")

    capsule_api = CapsuleApi(
        ctx.obj.api_url,
        ctx.obj.perimeter,
    )

    # First, find the capsule using list_and_filter_capsules
    filtered_capsules = list_and_filter_capsules(
        capsule_api, None, None, name, None, None, cap_id
    )

    if len(filtered_capsules) == 0:
        identifier = name if name else cap_id
        identifier_type = "name" if name else "id"
        raise AppConfigError(f"No app found with {identifier_type}: {identifier}")

    if len(filtered_capsules) > 1:
        raise AppConfigError(
            f"Multiple apps found with name: {name}. Please use --id to specify exactly which app you want logs for."
        )

    capsule = filtered_capsules[0]
    capsule_id = capsule.get("id")

    # Get workers
    try:
        workers_info = capsule_api.get_workers(capsule_id)
    except Exception as e:
        raise AppConfigError(f"Error retrieving workers for app {capsule_id}: {e}")

    if not workers_info:
        raise AppConfigError(f"No workers found for app {capsule_id}")

    # If worker_id not provided, show interactive selection
    if not worker_id:
        if len(workers_info) == 1:
            # Only one worker, use it automatically
            selected_worker = workers_info[0]
            worker_id = selected_worker.get("workerId")
            worker_phase = selected_worker.get("phase", "N/A")
            worker_version = selected_worker.get("version", "N/A")[:10]
            click.echo(
                f"📋 Using the only available worker: {worker_id[:20]}... (phase: {worker_phase}, version: {worker_version}...)"
            )
        else:
            # Multiple workers, show selection
            click.secho(
                "📋 Multiple workers found. Please select one:",
                fg=ColorTheme.INFO_COLOR,
                bold=True,
            )

            # Display workers in a table format for better readability
            headers = ["#", "Worker ID", "Phase", "Version", "Activity"]
            table_data = []

            for i, worker in enumerate(workers_info, 1):
                w_id = worker.get("workerId", "N/A")
                phase = worker.get("phase", "N/A")
                version = worker.get("version", "N/A")
                activity = str(worker.get("activity", "N/A"))

                table_data.append(
                    [
                        str(i),
                        w_id[:30] + "..." if len(w_id) > 33 else w_id,
                        phase,
                        version[:15] + "..." if len(version) > 18 else version,
                        activity,
                    ]
                )

            print_table(table_data, headers)

            # Create choices for the prompt
            worker_choices = []
            for i, worker in enumerate(workers_info, 1):
                worker_choices.append(str(i))

            selected_index = click.prompt(
                click.style(
                    "Select worker number", fg=ColorTheme.INFO_COLOR, bold=True
                ),
                type=click.Choice(worker_choices),
            )

            # Get the selected worker
            selected_worker = workers_info[int(selected_index) - 1]
            worker_id = selected_worker.get("workerId")

    # Get logs for the selected worker
    try:
        logs_response = capsule_api.logs(capsule_id, worker_id, previous=previous)
    except Exception as e:
        raise AppConfigError(f"Error retrieving logs for worker {worker_id}: {e}")

    # Format logs content
    logs_content = "\n".join([log.get("message", "") for log in logs_response])

    # Display or save logs
    if file:
        try:
            with open(file, "w") as f:
                f.write(logs_content)
            click.echo(f"📁 Logs saved to {file}")
        except Exception as e:
            raise AppConfigError(f"Error saving logs to file {file}: {e}")
    else:
        if logs_content.strip():
            click.echo(logs_content)
        else:
            click.echo("📝 No logs available for this worker.")


# if __name__ == "__main__":
#     cli()



"""
Examples demonstrating how to use the code packager abstraction.

This file provides usage examples for the code packager classes.
These examples are for documentation purposes and are not meant to be run directly.
"""

import os
import sys
from io import BytesIO
from typing import List, Dict, Any, Callable, Optional

from .code_packager import CodePackager


def basic_usage_example(datastore_type: str = "s3") -> None:
    """
    Basic usage example with ContentAddressedStore.

    This example shows how to:
    1. Define paths to include in a package
    2. Create a package using CodePackager's default packaging
    3. Store the package using ContentAddressedStore directly
    4. Generate a download command

    Parameters
    ----------
    datastore_type : str, default "s3"
        The type of datastore to use: "s3", "azure", "gs", or "local"
    """
    # Define the paths to include in the package
    paths_to_include = ["./"]

    # Define which file suffixes to include
    file_suffixes = [".py", ".md"]

    # Create metadata for the package
    metadata = {"example": True, "timestamp": "2023-01-01T00:00:00Z"}

    # Initialize the packager with datastore configuration
    packager = CodePackager(
        datastore_type=datastore_type,
        code_package_prefix="my-custom-prefix",  # Optional
    )

    # Store the package with packaging parameters
    package_url, package_key = packager.store(
        paths_to_include=paths_to_include,
        file_suffixes=file_suffixes,
        metadata=metadata,
    )

    # Generate a download command
    download_cmd = CodePackager.get_download_cmd(
        package_url=package_url,
        datastore_type=datastore_type,
        target_file="my_package.tar",
    )

    # Generate complete package commands for downloading and setup
    package_commands = packager.get_package_commands(
        code_package_url=package_url,
        target_file="my_package.tar",
        working_dir="my_app",
    )

    # Print some information
    print(f"Package URL: {package_url}")
    print(f"Package Key: {package_key}")
    print(f"Download Command: {download_cmd}")
    print(f"Complete package commands: {package_commands}")


def usage_with_custom_package_function(datastore_type: str = "s3") -> None:
    """
    Example of using the CodePackager with a custom package creation function.

    Parameters
    ----------
    datastore_type : str, default "s3"
        The type of datastore to use: "s3", "azure", "gs", or "local"
    """

    # Define a custom package function
    def create_custom_package():
        # This is a simple example - in real use, you might create a more complex package
        from io import BytesIO
        import tarfile
        import time

        buf = BytesIO()
        with tarfile.open(fileobj=buf, mode="w:gz") as tar:
            # Add a simple file to the tarball
            content = b"print('Hello, custom package!')"
            info = tarfile.TarInfo(name="hello.py")
            info.size = len(content)
            info.mtime = int(time.time())
            file_object = BytesIO(content)
            tar.addfile(info, file_object)

        return bytearray(buf.getvalue())

    # Initialize the packager with datastore configuration
    packager = CodePackager(datastore_type=datastore_type)

    # Store the package with the custom package function
    package_url, package_key = packager.store(package_create_fn=create_custom_package)

    # Generate a download command
    download_cmd = CodePackager.get_download_cmd(
        package_url=package_url,
        datastore_type=datastore_type,
        target_file="custom_package.tar",
    )

    # Generate complete package commands
    package_commands = packager.get_package_commands(
        code_package_url=package_url,
    )

    # Print some information
    print(f"Package URL: {package_url}")
    print(f"Package Key: {package_key}")
    print(f"Download Command: {download_cmd}")
    print(f"Complete commands: {package_commands}")



import os
import json
import tempfile
from contextlib import contextmanager


@contextmanager
def hide_access_keys(*args, **kwds):
    if "AWS_ACCESS_KEY_ID" in os.environ:
        AWS_ACCESS_KEY_ID = os.environ["AWS_ACCESS_KEY_ID"]
        del os.environ["AWS_ACCESS_KEY_ID"]
    else:
        AWS_ACCESS_KEY_ID = None
    if "AWS_SECRET_ACCESS_KEY" in os.environ:
        AWS_SECRET_ACCESS_KEY = os.environ["AWS_SECRET_ACCESS_KEY"]
        del os.environ["AWS_SECRET_ACCESS_KEY"]
    else:
        AWS_SECRET_ACCESS_KEY = None
    if "AWS_SESSION_TOKEN" in os.environ:
        AWS_SESSION_TOKEN = os.environ["AWS_SESSION_TOKEN"]
        del os.environ["AWS_SESSION_TOKEN"]
    else:
        AWS_SESSION_TOKEN = None
    try:
        yield
    finally:
        if AWS_ACCESS_KEY_ID is not None:
            os.environ["AWS_ACCESS_KEY_ID"] = AWS_ACCESS_KEY_ID
        if AWS_SECRET_ACCESS_KEY is not None:
            os.environ["AWS_SECRET_ACCESS_KEY"] = AWS_SECRET_ACCESS_KEY
        if AWS_SESSION_TOKEN is not None:
            os.environ["AWS_SESSION_TOKEN"] = AWS_SESSION_TOKEN


# This is a special placeholder value that can be passed as role_arn to
# get_boto3_session() which makes it use the CSPR role, if its set.
USE_CSPR_ROLE_ARN_IF_SET = "__cspr__"


def get_boto3_session(role_arn=None, session_vars=None):
    import boto3
    import botocore
    from metaflow_extensions.outerbounds.plugins.auth_server import get_token
    from metaflow_extensions.outerbounds.plugins.aws.assume_role import (
        OBP_ASSUME_ROLE_ARN_ENV_VAR,
    )

    from hashlib import sha256
    from metaflow.util import get_username

    user = get_username()

    token_info = get_token("/generate/aws")

    # Write token to a file. The file name is derived from the user name
    # so it works with multiple users on the same machine.
    #
    # We hash the user name so we don't have to deal with special characters
    # in the file name and the file name is not exposed to the user
    # anyways, so it doesn't matter that its a little ugly.
    token_file = "/tmp/obp_token." + sha256(user.encode("utf-8")).hexdigest()[:16]

    # Write to a temp file then rename to avoid a situation when someone
    # tries to read the file after it was open for writing (and truncated)
    # but before the token was written to it.
    with tempfile.NamedTemporaryFile("w", delete=False) as f:
        f.write(token_info["token"])
        tmp_token_file = f.name
    os.rename(tmp_token_file, token_file)

    cspr_role = None
    if token_info.get("cspr_role_arn"):
        cspr_role = token_info["cspr_role_arn"]

    # Check if the assume_role decorator has set a CSPR ARN via environment variable
    # This takes precedence over CSPR role that comes from the token_info response
    decorator_role_arn = os.environ.get(OBP_ASSUME_ROLE_ARN_ENV_VAR)
    if decorator_role_arn:
        cspr_role = decorator_role_arn

    if cspr_role:
        # If CSPR role is set, we set it as the default role to assume
        # for the AWS SDK. We do this by writing an AWS config file
        # with two profiles. One to get credentials for the task role
        # in exchange for the OIDC token, and second to assume the
        # CSPR role using the task role credentials.
        import configparser
        from io import StringIO

        aws_config = configparser.ConfigParser()

        # Task role profile
        aws_config["profile task"] = {
            "role_arn": token_info["role_arn"],
            "web_identity_token_file": token_file,
        }

        # CSPR role profile (default)
        aws_config["profile cspr"] = {
            "role_arn": cspr_role,
            "source_profile": "task",
        }

        aws_config_string = StringIO()
        aws_config.write(aws_config_string)
        aws_config_file = (
            "/tmp/aws_config." + sha256(user.encode("utf-8")).hexdigest()[:16]
        )
        with tempfile.NamedTemporaryFile(
            "w", delete=False, dir=os.path.dirname(aws_config_file)
        ) as f:
            f.write(aws_config_string.getvalue())
            tmp_aws_config_file = f.name
        os.rename(tmp_aws_config_file, aws_config_file)
        os.environ["AWS_CONFIG_FILE"] = aws_config_file
        os.environ["AWS_PROFILE"] = "cspr"
    else:
        os.environ["AWS_WEB_IDENTITY_TOKEN_FILE"] = token_file
        os.environ["AWS_ROLE_ARN"] = token_info["role_arn"]

    # Enable regional STS endpoints. This is the new recommended way
    # by AWS [1] and is the more performant way.
    # [1] https://docs.aws.amazon.com/sdkref/latest/guide/feature-sts-regionalized-endpoints.html
    os.environ["AWS_STS_REGIONAL_ENDPOINTS"] = "regional"
    if token_info.get("region"):
        os.environ["AWS_DEFAULT_REGION"] = token_info["region"]

    if cspr_role:
        #  The generated AWS config will be used here since we set the
        #  AWS_CONFIG_FILE environment variable above.
        if role_arn == USE_CSPR_ROLE_ARN_IF_SET:
            # Otherwise start from the default profile, assuming CSPR role
            session = boto3.session.Session(profile_name="cspr")
        else:
            session = boto3.session.Session(profile_name="task")
    else:
        # Not using AWS config, just AWS_WEB_IDENTITY_TOKEN_FILE + AWS_ROLE_ARN
        session = boto3.session.Session()

    if role_arn and role_arn != USE_CSPR_ROLE_ARN_IF_SET:
        # If the user provided a role_arn, we assume that role
        # using the task role credentials. CSPR role is not used.
        fetcher = botocore.credentials.AssumeRoleCredentialFetcher(
            client_creator=session._session.create_client,
            source_credentials=session._session.get_credentials(),
            role_arn=role_arn,
            extra_args={},
        )
        creds = botocore.credentials.DeferredRefreshableCredentials(
            method="assume-role", refresh_using=fetcher.fetch_credentials
        )
        botocore_session = botocore.session.Session(session_vars=session_vars)
        botocore_session._credentials = creds
        return boto3.session.Session(botocore_session=botocore_session)
    else:
        # If the user didn't provide a role_arn, or if the role_arn
        # is set to USE_CSPR_ROLE_ARN_IF_SET, we return the default
        # session which would use the CSPR role if it is set on the
        # server, and the task role otherwise.
        return session


class ObpAuthProvider(object):
    name = "obp"

    @staticmethod
    def get_client(
        module, with_error=False, role_arn=None, session_vars=None, client_params=None
    ):
        if client_params is None:
            client_params = {}

        from botocore.exceptions import ClientError
        from botocore.config import Config

        with hide_access_keys():
            session = get_boto3_session(role_arn, session_vars)
            _client_params = client_params.copy()
            if _client_params.get("config") and type(_client_params["config"]) == dict:
                _client_params["config"] = Config(**_client_params["config"])

            if with_error:
                return session.client(module, **_client_params), ClientError
            else:
                return session.client(module, **_client_params)


AWS_CLIENT_PROVIDERS_DESC = [("obp", ".ObpAuthProvider")]


class ObpAzureAuthProvider(object):
    name = "obp"

    @staticmethod
    def create_cacheable_azure_credential(*args, **kwargs):
        """azure.identity.DefaultAzureCredential is not readily cacheable in a dictionary
        because it does not have a content based hash and equality implementations.

        We implement a subclass CacheableDefaultAzureCredential to add them.

        We need this because credentials will be part of the cache key in _ClientCache.
        """
        from azure.identity import WorkloadIdentityCredential

        from metaflow_extensions.outerbounds.plugins.auth_server import get_token

        class CacheableDefaultAzureCredential(WorkloadIdentityCredential):
            def __init__(self, *args, **kwargs):
                super(CacheableDefaultAzureCredential, self).__init__(*args, **kwargs)
                # Just hashing all the kwargs works because they are all individually
                # hashable as of 7/15/2022.
                #
                # What if Azure adds unhashable things to kwargs?
                # - We will have CI to catch this (it will always install the latest Azure SDKs)
                # - In Metaflow usage today we never specify any kwargs anyway. (see last line
                #   of the outer function.
                self._hash_code = hash((args, tuple(sorted(kwargs.items()))))

            def __hash__(self):
                return self._hash_code

            def __eq__(self, other):
                return hash(self) == hash(other)

        from hashlib import sha256
        from metaflow.util import get_username

        user = get_username()

        token_info = get_token("/generate/azure")
        token_file = "/tmp/obp_token." + sha256(user.encode("utf-8")).hexdigest()[:16]

        with tempfile.NamedTemporaryFile("w", delete=False) as f:
            f.write(token_info["token"])
            tmp_token_file = f.name
        os.rename(tmp_token_file, token_file)
        return CacheableDefaultAzureCredential(
            tenant_id=token_info["azureTenantId"],
            client_id=token_info["azureClientId"],
            token_file_path=token_file,
        )


AZURE_CLIENT_PROVIDERS_DESC = [("obp", ".ObpAzureAuthProvider")]

import threading
import time

_gcp_client_cache = dict()


def _get_cache_key():
    return os.getpid(), threading.get_ident()


class ObpGcpAuthProvider(object):
    name = "obp"

    @staticmethod
    def get_gs_storage_client(*args, **kwargs):
        import sys
        from metaflow_extensions.outerbounds.plugins.auth_server import get_token

        cache_key = _get_cache_key()
        if _gcp_client_cache.get(cache_key):
            # Don't cache the client for more than 5 minutes as it may have
            # expired.
            if _gcp_client_cache[cache_key]._created_at < time.time() - 300:
                del _gcp_client_cache[cache_key]
            else:
                return _gcp_client_cache[cache_key]

        from hashlib import sha256
        from metaflow.util import get_username

        user = get_username()

        token_info = get_token("/generate/gcp")
        token_file = "/tmp/obp_token." + sha256(user.encode("utf-8")).hexdigest()[:16]
        credentials_file = (
            "/tmp/obp_credentials." + sha256(user.encode("utf-8")).hexdigest()[:16]
        )

        with tempfile.NamedTemporaryFile("w", delete=False) as f:
            f.write(token_info["token"])
            tmp_token_file = f.name
        os.rename(tmp_token_file, token_file)

        credentials_json = {
            "type": "external_account",
            "audience": f"//iam.googleapis.com/projects/{token_info['gcpProjectNumber']}/locations/global/workloadIdentityPools/{token_info['gcpWorkloadIdentityPool']}/providers/{token_info['gcpWorkloadIdentityPoolProvider']}",
            "subject_token_type": "urn:ietf:params:oauth:token-type:jwt",
            "token_url": "https://sts.googleapis.com/v1/token",
            "service_account_impersonation_url": f"https://iamcredentials.googleapis.com/v1/projects/-/serviceAccounts/{token_info['gcpServiceAccountEmail']}:generateAccessToken",
            "credential_source": {
                "file": token_file,
                "format": {"type": "text"},
            },
        }

        with tempfile.NamedTemporaryFile("w", delete=False) as f:
            f.write(json.dumps(credentials_json))
            tmp_credentials_file = f.name
        os.rename(tmp_credentials_file, credentials_file)

        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = credentials_file
        from google.cloud import storage

        storage_client = storage.Client(project=token_info["gcpProjectId"])
        storage_client._created_at = time.time()
        _gcp_client_cache[cache_key] = storage_client
        return storage_client

    @staticmethod
    def get_credentials(scopes, *args, **kwargs):
        import google.auth

        return google.auth.default(scopes=scopes)


GCP_CLIENT_PROVIDERS_DESC = [("obp", ".ObpGcpAuthProvider")]
CLIS_DESC = [
    ("nvidia", ".nvcf.nvcf_cli.cli"),
    ("nvct", ".nvct.nvct_cli.cli"),
    ("fast-bakery", ".fast_bakery.fast_bakery_cli.cli"),
    ("snowpark", ".snowpark.snowpark_cli.cli"),
]
STEP_DECORATORS_DESC = [
    ("nvidia", ".nvcf.nvcf_decorator.NvcfDecorator"),
    ("nvct", ".nvct.nvct_decorator.NvctDecorator"),
    (
        "fast_bakery_internal",
        ".fast_bakery.fast_bakery_decorator.InternalFastBakeryDecorator",
    ),
    ("snowpark", ".snowpark.snowpark_decorator.SnowparkDecorator"),
    ("tensorboard", ".tensorboard.TensorboardDecorator"),
    ("gpu_profile", ".profilers.gpu_profile_decorator.GPUProfileDecorator"),
    ("test_append_card", ".profilers.simple_card_decorator.DynamicCardAppendDecorator"),
    ("nim", ".nim.nim_decorator.NimDecorator"),
    ("ollama", ".ollama.OllamaDecorator"),
    ("vllm", ".vllm.VLLMDecorator"),
    ("app_deploy", ".apps.app_deploy_decorator.AppDeployDecorator"),
    ("s3_proxy", ".s3_proxy.s3_proxy_decorator.S3ProxyDecorator"),
    ("nebius_s3_proxy", ".s3_proxy.s3_proxy_decorator.NebiusS3ProxyDecorator"),
    ("coreweave_s3_proxy", ".s3_proxy.s3_proxy_decorator.CoreWeaveS3ProxyDecorator"),
]

TOGGLE_STEP_DECORATOR = [
    "-batch",
    "-step_functions_internal",
    "-airflow_internal",
]

TOGGLE_CLI = ["-batch", "-step-functions", "-airflow"]

ENVIRONMENTS_DESC = [
    ("fast-bakery", ".fast_bakery.docker_environment.DockerEnvironment")
]

SECRETS_PROVIDERS_DESC = [
    ("outerbounds", ".secrets.secrets.OuterboundsSecretsProvider"),
]
# Adding an override here so the library can be imported at the metaflow.plugins level
__mf_promote_submodules__ = ["snowflake", "ollama", "torchtune", "optuna"]



import os
import sys
import time
import tarfile
import json
from io import BytesIO
from typing import List, Tuple, Dict, Any, Optional, Callable, Union

from metaflow.datastore.content_addressed_store import ContentAddressedStore
from metaflow.util import to_unicode
from metaflow.metaflow_config import (
    DATASTORE_SYSROOT_S3,
    DATASTORE_SYSROOT_AZURE,
    DATASTORE_SYSROOT_GS,
    DATASTORE_SYSROOT_LOCAL,
)

DEFAULT_FILE_SUFFIXES = [
    ".py",
    ".txt",
    ".yaml",
    ".yml",
    ".json",
    ".html",
    ".css",
    ".js",
    ".jsx",
    ".ts",
    ".tsx",
    ".md",
    ".rst",
]
# Default prefix for code packages in content addressed store
CODE_PACKAGE_PREFIX = "apps-code-packages"


# this is os.walk(follow_symlinks=True) with cycle detection
def walk_without_cycles(top_root):
    seen = set()

    def _recurse(root):
        for parent, dirs, files in os.walk(root):
            for d in dirs:
                path = os.path.join(parent, d)
                if os.path.islink(path):
                    # Breaking loops: never follow the same symlink twice
                    #
                    # NOTE: this also means that links to sibling links are
                    # not followed. In this case:
                    #
                    #   x -> y
                    #   y -> oo
                    #   oo/real_file
                    #
                    # real_file is only included twice, not three times
                    reallink = os.path.realpath(path)
                    if reallink not in seen:
                        seen.add(reallink)
                        for x in _recurse(path):
                            yield x
            yield parent, files

    for x in _recurse(top_root):
        yield x


def symlink_friendly_walk(root, exclude_hidden=True, suffixes=None):
    if suffixes is None:
        suffixes = []
    root = to_unicode(root)  # handle files/folder with non ascii chars
    prefixlen = len("%s/" % os.path.dirname(root))
    for (
        path,
        files,
    ) in walk_without_cycles(root):
        if exclude_hidden and "/." in path:
            continue
        # path = path[2:] # strip the ./ prefix
        # if path and (path[0] == '.' or './' in path):
        #    continue
        for fname in files:
            if (fname[0] == "." and fname in suffixes) or (
                fname[0] != "." and any(fname.endswith(suffix) for suffix in suffixes)
            ):
                p = os.path.join(path, fname)
                yield p, p[prefixlen:]


class CodePackager:
    """
    A datastore-agnostic class for packaging code.

    This class handles creating a code package (tarball) for deployment
    and provides methods for storing and retrieving it using Metaflow's
    ContentAddressedStore directly.

    Usage examples:
    ```python
    packager = CodePackager(
        datastore_type: str = "s3",
        datastore_root = None,
        code_package_prefix = None,
    )

    package_url, package_key = packager.store(
        paths_to_include = ["./"],
        file_suffixes = [".py", ".txt", ".yaml", ".yml", ".json"],
    )

    package_url, package_key = packager.store(
        package_create_fn = lambda: my_custom_package_create_fn(),
    )
    ```
    """

    def __init__(
        self,
        datastore_type: str = "s3",
        datastore_root: Optional[str] = None,
        code_package_prefix: Optional[str] = None,
    ):
        """
        Initialize the CodePackager with datastore configuration.

        Parameters
        ----------
        datastore_type : str, default "s3"
            The type of datastore to use: "s3", "azure", "gs", or "local"
        datastore_root : str, optional
            Root path for the datastore. If not provided, uses the default for the datastore type.
        code_package_prefix : str, optional
            The prefix to use for storing code packages in the content addressed store.
            If not provided, uses the CODE_PACKAGE_PREFIX configuration value.
        """
        self._datastore_type = datastore_type
        self._datastore_root = datastore_root
        self._code_package_prefix = code_package_prefix

    def store(
        self,
        package_create_fn: Optional[Callable[[], bytes]] = None,
        paths_to_include: Optional[List[str]] = None,
        file_suffixes: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Tuple[str, str]:
        """
        Create and store a code package using Metaflow's ContentAddressedStore.

        This method can be called in two ways:
        1. With paths_to_include and file_suffixes to use the default packaging
        2. With a custom package_create_fn for custom packaging logic

        Parameters
        ----------
        package_create_fn : Callable[[], bytes], optional
            A function that creates and returns a package as bytes.
            This allows for custom packaging logic without dependency on specific objects.
        paths_to_include : List[str], optional
            List of paths to include in the package. Used by default_package_create.
        file_suffixes : List[str], optional
            List of file suffixes to include. Used by default_package_create.
        metadata : Dict[str, Any], optional
            Metadata to include in the package when using default_package_create.

        Returns
        -------
        Tuple[str, str]
            A tuple containing (package_url, package_key) that identifies the location
            and content-addressed key of the stored package.
        """
        # Prepare default values
        _paths_to_include = paths_to_include or []
        _file_suffixes = file_suffixes or DEFAULT_FILE_SUFFIXES
        _metadata = metadata or {}

        # If no package_create_fn provided, use default_package_create
        if package_create_fn is None:
            _package_create_fn = lambda: self.default_package_create(
                _paths_to_include, _file_suffixes, _metadata
            )
        else:
            _package_create_fn = package_create_fn

        # Create the package
        code_package = _package_create_fn()

        # Get the ContentAddressedStore for the specified datastore
        ca_store = self.get_content_addressed_store(
            datastore_type=self._datastore_type,
            datastore_root=self._datastore_root,
            prefix=(
                str(self._code_package_prefix)
                if self._code_package_prefix is not None
                else str(CODE_PACKAGE_PREFIX)
            ),
        )

        # Store the package using raw=True to ensure we can access it directly via URL
        results = ca_store.save_blobs([code_package], raw=True, len_hint=1)
        package_url, package_key = results[0].uri, results[0].key

        return package_url, package_key

    @staticmethod
    def get_content_addressed_store(
        datastore_type: str = "s3",
        datastore_root: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> ContentAddressedStore:
        """
        Get a ContentAddressedStore instance for the specified datastore.

        Parameters
        ----------
        datastore_type : str, default "s3"
            Type of datastore: "s3", "azure", "gs", or "local"
        datastore_root : str, optional
            Root path for the datastore. If not provided, uses the default for the datastore type.
        prefix : str, optional
            Prefix to use when storing objects in the datastore.
            If not provided, uses the CODE_PACKAGE_PREFIX configuration value.

        Returns
        -------
        ContentAddressedStore
            A ContentAddressedStore instance configured for the specified datastore
        """
        from metaflow.plugins import DATASTORES

        datastore_impls = [i for i in DATASTORES if i.TYPE == datastore_type]
        if len(datastore_impls) == 0:
            raise ValueError(f"Unsupported datastore type: {datastore_type}")
        if len(datastore_impls) > 1:
            raise ValueError(
                f"Multiple datastore implementations found for type: {datastore_type}"
            )
        datastore_impl = datastore_impls[0]
        root = None
        # Import the storage implementation based on datastore_type
        if datastore_type == "s3":
            root = datastore_root or DATASTORE_SYSROOT_S3
        elif datastore_type == "azure":
            root = datastore_root or DATASTORE_SYSROOT_AZURE
        elif datastore_type == "gs":
            root = datastore_root or DATASTORE_SYSROOT_GS
        elif datastore_type == "local":
            root = datastore_root or DATASTORE_SYSROOT_LOCAL

        # Ensure prefix is a string
        store_prefix = str(prefix) if prefix is not None else str(CODE_PACKAGE_PREFIX)

        storage_impl = datastore_impl(root=root)
        # Create and return a ContentAddressedStore
        return ContentAddressedStore(prefix=store_prefix, storage_impl=storage_impl)

    @staticmethod
    def get_download_cmd(
        package_url: str,
        datastore_type: str,
        python_cmd: str = "python",
        target_file: str = "job.tar",
        escape_quotes: bool = True,
    ) -> str:
        """
        Generate a command to download the code package.

        Parameters
        ----------
        package_url : str
            The URL of the package to download
        datastore_type : str
            The type of datastore (s3, azure, gs, local)
        python_cmd : str, optional
            The Python command to use
        target_file : str, optional
            The target file name to save the package as
        escape_quotes : bool, optional
            Whether to escape quotes in the command

        Returns
        -------
        str
            A shell command string to download the package
        """
        if datastore_type == "s3":
            from metaflow.plugins.aws.aws_utils import parse_s3_full_path

            bucket, s3_object = parse_s3_full_path(package_url)
            # Simplify the script and use single quotes to avoid shell escaping issues
            script = 'import boto3, os; ep=os.getenv({quote}METAFLOW_S3_ENDPOINT_URL{quote}); boto3.client("s3", **({{"endpoint_url":ep}} if ep else {{}})).download_file({quote}{bucket}{quote}, {quote}{s3_object}{quote}, {quote}{target_file}{quote})'.format(
                quote='\\"' if escape_quotes else '"',
                bucket=bucket,
                s3_object=s3_object,
                target_file=target_file,
            )
            # Format the command with proper quoting
            return f"{python_cmd} -c '{script}'"
        elif datastore_type == "azure":
            from metaflow.plugins.azure.azure_utils import parse_azure_full_path

            container_name, blob = parse_azure_full_path(package_url)
            # remove a trailing slash, if present
            blob_endpoint = "${METAFLOW_AZURE_STORAGE_BLOB_SERVICE_ENDPOINT%/}"
            return "download-azure-blob --blob-endpoint={blob_endpoint} --container={container} --blob={blob} --output-file={target}".format(
                blob_endpoint=blob_endpoint,
                blob=blob,
                container=container_name,
                target=target_file,
            )
        elif datastore_type == "gs":
            from metaflow.plugins.gcp.gs_utils import parse_gs_full_path

            bucket_name, gs_object = parse_gs_full_path(package_url)
            return "download-gcp-object --bucket=%s --object=%s --output-file=%s" % (
                bucket_name,
                gs_object,
                target_file,
            )
        elif datastore_type == "local":
            # For local storage, simply copy the file
            return "cp %s %s" % (package_url, target_file)
        else:
            raise NotImplementedError(
                f"Download command not implemented for datastore type: {datastore_type}"
            )

    def get_package_commands(
        self,
        code_package_url: str,
        python_cmd: str = "python",
        target_file: str = "job.tar",
        working_dir: str = "metaflow",
        retries: int = 5,
        escape_quotes: bool = True,
    ) -> List[str]:
        """
        Get a complete list of shell commands to download and extract a code package.

        This method generates a comprehensive set of shell commands for downloading
        and extracting a code package, similar to MetaflowEnvironment.get_package_commands.

        Parameters
        ----------
        code_package_url : str
            The URL of the code package to download
        python_cmd : str, optional
            The Python command to use
        target_file : str, optional
            The target file name to save the package as
        working_dir : str, optional
            The directory to create and extract the package into
        retries : int, optional
            Number of download retries to attempt
        escape_quotes : bool, optional
            Whether to escape quotes in the command

        Returns
        -------
        List[str]
            List of shell commands to execute
        """
        # Use the datastore_type from initialization if not provided
        datastore_type = self._datastore_type

        # Helper function to create dependency installation command
        def _get_install_dependencies_cmd():
            base_cmd = "{} -m pip install -qqq --no-compile --no-cache-dir --disable-pip-version-check".format(
                python_cmd
            )

            datastore_packages = {
                "s3": ["boto3"],
                "azure": [
                    "azure-identity",
                    "azure-storage-blob",
                    "azure-keyvault-secrets",
                    "simple-azure-blob-downloader",
                ],
                "gs": [
                    "google-cloud-storage",
                    "google-auth",
                    "simple-gcp-object-downloader",
                    "google-cloud-secret-manager",
                ],
                "local": [],
            }

            if datastore_type not in datastore_packages:
                raise NotImplementedError(
                    "Unknown datastore type: {}".format(datastore_type)
                )

            if not datastore_packages[datastore_type]:
                return "# No dependencies required for local datastore"

            cmd = "{} {}".format(
                base_cmd, " ".join(datastore_packages[datastore_type] + ["requests"])
            )
            # Skip pip installs if we know packages might already be available
            return "if [ -z $METAFLOW_SKIP_INSTALL_DEPENDENCIES ]; then {}; fi".format(
                cmd
            )

        download_cmd = self.get_download_cmd(
            code_package_url, datastore_type, python_cmd, target_file, escape_quotes
        )

        # Define the log functions for bash
        bash_mflog = (
            'function mflog() { echo "[$(date -u +"%Y-%m-%dT%H:%M:%SZ")]" "$@"; }'
        )
        bash_flush_logs = 'function flush_mflogs() { echo "[$(date -u +"%Y-%m-%dT%H:%M:%SZ")] Flushing logs"; }'

        cmds = [
            bash_mflog,
            bash_flush_logs,
            "mflog 'Setting up task environment.'",
            _get_install_dependencies_cmd(),
            f"mkdir -p {working_dir}",
            f"cd {working_dir}",
            "mkdir -p .metaflow",  # mute local datastore creation log
            f"i=0; while [ $i -le {retries} ]; do "
            "mflog 'Downloading code package...'; "
            + download_cmd
            + " && mflog 'Code package downloaded.' && break; "
            "sleep 10; i=$((i+1)); "
            "done",
            f"if [ $i -gt {retries} ]; then "
            "mflog 'Failed to download code package from %s "
            f"after {retries+1} tries. Exiting...' && exit 1; "
            "fi" % code_package_url,
            "TAR_OPTIONS='--warning=no-timestamp' tar xf %s" % target_file,
            "mflog 'Task is starting.'",
            "flush_mflogs",
        ]

        return cmds

    @staticmethod
    def directory_walker(
        root,
        exclude_hidden=True,
        suffixes=None,
    ) -> List[Tuple[str, str]]:
        """
        Walk a directory and yield tuples of (file_path, relative_arcname) for files
        that match the given suffix filters. It will follow symlinks, but not cycles.

        This function is similar to MetaflowPackage._walk and handles symlinks safely.

        Parameters
        ----------
        root : str
            The root directory to walk
        exclude_hidden : bool, default True
            Whether to exclude hidden files and directories (those starting with '.')
        suffixes : List[str], optional
            List of file suffixes to include (e.g. ['.py', '.txt'])

        Returns
        -------
        List[Tuple[str, str]]
            List of tuples (file_path, relative_arcname) where:
            - file_path is the full path to the file
            - relative_arcname is the path to use within the archive
        """
        files = []
        for file_path, rel_path in symlink_friendly_walk(
            root, exclude_hidden, suffixes
        ):
            files.append((file_path, rel_path))
        return files

    @staticmethod
    def default_package_create(
        paths: List[str], suffixes: List[str], metadata: Optional[Dict[str, Any]] = None
    ) -> bytes:
        """
        Create a default tarball package from specified paths.

        Parameters
        ----------
        paths : List[str]
            List of paths to include in the package
        suffixes : List[str]
            List of file suffixes to include
        metadata : Dict[str, Any], optional
            Metadata to include in the package

        Returns
        -------
        bytes
            The binary content of the tarball
        """
        buf = BytesIO()

        with tarfile.open(fileobj=buf, mode="w:gz", compresslevel=3) as tar:
            # Add metadata if provided
            if metadata:
                metadata_buf = BytesIO()
                metadata_buf.write(json.dumps(metadata).encode("utf-8"))
                metadata_buf.seek(0)
                info = tarfile.TarInfo("metadata.json")
                info.size = len(metadata_buf.getvalue())
                info.mtime = 1747158696  # 13 May 2025 10:31:36 (so that we dont have a changing hash everytime we run)
                tar.addfile(info, metadata_buf)

            def no_mtime(tarinfo):
                # a modification time change should not change the hash of
                # the package. Only content modifications will.
                # Setting this default to Dec 3, 2019
                tarinfo.mtime = 1747158696  # 13 May 2025 10:31:36 (so that we dont have a changing hash everytime we run)
                return tarinfo

            # Add files from specified paths
            for path in paths:
                if os.path.isdir(path):
                    # Use directory_walker for directories to handle symlinks properly
                    for file_path, rel_path in CodePackager.directory_walker(
                        path,
                        exclude_hidden=True,
                        suffixes=suffixes,
                    ):
                        tar.add(
                            file_path,
                            arcname=rel_path,
                            filter=no_mtime,
                            recursive=False,
                        )
                elif os.path.isfile(path):
                    if any(path.endswith(suffix) for suffix in suffixes):
                        tar.add(path, arcname=os.path.basename(path))

        tarball = bytearray(buf.getvalue())
        tarball[4:8] = [0] * 4  # Reset 4 bytes from offset 4 to account for ts
        return tarball

    @staticmethod
    def _add_tar_file(tar, filename, buf):
        tarinfo = tarfile.TarInfo(name=filename)
        tarinfo.size = len(buf.getvalue())
        buf.seek(0)
        tarinfo.mtime = 1747158696  # 13 May 2025 10:31:36 (so that we dont have a changing hash everytime we run)
        tar.addfile(tarinfo, fileobj=buf)

    @classmethod
    def package_directory(
        cls,
        directory_path: str,
        suffixes: Optional[List[str]] = None,
        exclude_hidden: bool = True,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> bytes:
        """
        Package a directory and all of its contents that match the given suffixes.

        This is a convenience method that works similarly to MetaflowPackage._walk
        to package a directory for deployment. Will default follow_symlinks.

        Parameters
        ----------
        directory_path : str
            The directory to package
        suffixes : List[str], optional
            List of file suffixes to include (defaults to standard code extensions)
        exclude_hidden : bool, default True
            Whether to exclude hidden files and directories
        metadata : Dict[str, Any], optional
            Metadata to include in the package
        Returns
        -------
        bytes
            The binary content of the tarball
        """
        if not os.path.isdir(directory_path):
            raise ValueError(f"The path '{directory_path}' is not a directory")

        # Use default suffixes if none provided
        if suffixes is None:
            suffixes = [".py", ".txt", ".yaml", ".yml", ".json"]

        buf = BytesIO()

        def no_mtime(tarinfo):
            # a modification time change should not change the hash of
            # the package. Only content modifications will.
            # Setting this to a fixed date so that we don't have a changing hash everytime we run
            tarinfo.mtime = 1747158696  # 13 May 2025 10:31:36
            return tarinfo

        with tarfile.open(
            fileobj=buf, mode="w:gz", compresslevel=3, dereference=True
        ) as tar:
            # Add metadata if provided
            if metadata:
                cls._add_tar_file(
                    tar, "metadata.json", BytesIO(json.dumps(metadata).encode("utf-8"))
                )

            # Walk the directory and add matching files
            for file_path, rel_path in cls.directory_walker(
                directory_path,
                exclude_hidden=exclude_hidden,
                suffixes=suffixes,
            ):
                # Remove debug print statement
                tar.add(file_path, arcname=rel_path, recursive=False, filter=no_mtime)

        tarball = bytearray(buf.getvalue())
        tarball[4:8] = [0] * 4  # Reset 4 bytes from offset 4 to account for ts
        return tarball



from metaflow.decorators import StepDecorator
from metaflow import current
import functools
import os
import threading

from .ollama import OllamaManager, OllamaRequestInterceptor
from .status_card import OllamaStatusCard
from ..card_utilities.injector import CardDecoratorInjector

__mf_promote_submodules__ = ["plugins.ollama"]


class OllamaDecorator(StepDecorator, CardDecoratorInjector):
    """
    This decorator is used to run Ollama APIs as Metaflow task sidecars.

    User code call
    --------------
    @ollama(
        models=[...],
        ...
    )

    Valid backend options
    ---------------------
    - 'local': Run as a separate process on the local task machine.
    - (TODO) 'managed': Outerbounds hosts and selects compute provider.
    - (TODO) 'remote': Spin up separate instance to serve Ollama models.

    Valid model options
    -------------------
    Any model here https://ollama.com/search, e.g. 'llama3.2', 'llama3.3'

    Parameters
    ----------
    models: list[str]
        List of Ollama containers running models in sidecars.
    backend: str
        Determines where and how to run the Ollama process.
    force_pull: bool
        Whether to run `ollama pull` no matter what, or first check the remote cache in Metaflow datastore for this model key.
    cache_update_policy: str
        Cache update policy: "auto", "force", or "never".
    force_cache_update: bool
        Simple override for "force" cache update policy.
    debug: bool
        Whether to turn on verbose debugging logs.
    circuit_breaker_config: dict
        Configuration for circuit breaker protection. Keys: failure_threshold, recovery_timeout, reset_timeout.
    timeout_config: dict
        Configuration for various operation timeouts. Keys: pull, stop, health_check, install, server_startup.
    """

    name = "ollama"
    defaults = {
        "models": [],
        "backend": "local",
        "force_pull": False,
        "cache_update_policy": "auto",  # "auto", "force", "never"
        "force_cache_update": False,  # Simple override for "force"
        "debug": False,
        "circuit_breaker_config": {
            "failure_threshold": 3,
            "recovery_timeout": 60,
            "reset_timeout": 30,
        },
        "timeout_config": {
            "pull": 600,  # 10 minutes for model pulls
            "stop": 30,  # 30 seconds for model stops
            "health_check": 5,  # 5 seconds for health checks
            "install": 60,  # 1 minute for Ollama installation
            "server_startup": 300,  # 5 minutes for server startup
        },
        "card_refresh_interval": 10,  # seconds - how often to update the status card
    }

    def step_init(
        self, flow, graph, step_name, decorators, environment, flow_datastore, logger
    ):
        super().step_init(
            flow, graph, step_name, decorators, environment, flow_datastore, logger
        )
        self.flow_datastore_backend = flow_datastore._storage_impl

        # Attach the ollama status card
        self.attach_card_decorator(
            flow,
            step_name,
            "ollama_status",
            "blank",
            refresh_interval=self.attributes["card_refresh_interval"],
        )

    def task_decorate(
        self, step_func, flow, graph, retry_count, max_user_code_retries, ubf_context
    ):
        @functools.wraps(step_func)
        def ollama_wrapper():
            self.ollama_manager = None
            self.request_interceptor = None
            self.status_card = None
            self.card_monitor_thread = None

            try:
                # Initialize status card and monitoring
                self.status_card = OllamaStatusCard(
                    refresh_interval=self.attributes["card_refresh_interval"]
                )

                # Start card monitoring in background
                def monitor_card():
                    try:
                        self.status_card.on_startup(current.card["ollama_status"])

                        while not getattr(
                            self.card_monitor_thread, "_stop_event", False
                        ):
                            try:
                                # Trigger card update with current data
                                self.status_card.on_update(
                                    current.card["ollama_status"], None
                                )
                                import time

                                time.sleep(self.attributes["card_refresh_interval"])
                            except Exception as e:
                                if self.attributes["debug"]:
                                    print(f"[@ollama] Card monitoring error: {e}")
                                break
                    except Exception as e:
                        if self.attributes["debug"]:
                            print(f"[@ollama] Card monitor thread error: {e}")
                        self.status_card.on_error(current.card["ollama_status"], str(e))

                self.card_monitor_thread = threading.Thread(
                    target=monitor_card, daemon=True
                )
                self.card_monitor_thread._stop_event = False
                self.card_monitor_thread.start()

                # Initialize OllamaManager with status card
                self.ollama_manager = OllamaManager(
                    models=self.attributes["models"],
                    backend=self.attributes["backend"],
                    flow_datastore_backend=self.flow_datastore_backend,
                    force_pull=self.attributes["force_pull"],
                    cache_update_policy=self.attributes["cache_update_policy"],
                    force_cache_update=self.attributes["force_cache_update"],
                    debug=self.attributes["debug"],
                    circuit_breaker_config=self.attributes["circuit_breaker_config"],
                    timeout_config=self.attributes["timeout_config"],
                    status_card=self.status_card,
                )

                # Install request protection by monkey-patching ollama package
                self.request_interceptor = OllamaRequestInterceptor(
                    self.ollama_manager.circuit_breaker, self.attributes["debug"]
                )
                self.request_interceptor.install_protection()

                if self.attributes["debug"]:
                    print(
                        "[@ollama] OllamaManager initialized and request protection installed"
                    )

            except Exception as e:
                if self.status_card:
                    self.status_card.add_event(
                        "error", f"Initialization failed: {str(e)}"
                    )
                    try:
                        self.status_card.on_error(current.card["ollama_status"], str(e))
                    except:
                        pass
                print(f"[@ollama] Error initializing OllamaManager: {e}")
                raise

            try:
                if self.status_card:
                    self.status_card.add_event("info", "Starting user step function")
                step_func()
                if self.status_card:
                    self.status_card.add_event(
                        "success", "User step function completed successfully"
                    )
            finally:
                # Remove request protection first (before terminating models)
                if self.request_interceptor:
                    self.request_interceptor.remove_protection()
                    if self.attributes["debug"]:
                        print("[@ollama] Request protection removed")

                # Then cleanup ollama manager (while card monitoring is still active)
                if self.ollama_manager:
                    self.ollama_manager.terminate_models()

                # Give the card a moment to render the final shutdown events
                if self.card_monitor_thread and self.status_card:
                    import time

                    # Trigger one final card update to capture all shutdown events
                    try:
                        self.status_card.on_update(current.card["ollama_status"], None)
                    except Exception as e:
                        if self.attributes["debug"]:
                            print(f"[@ollama] Final card update error: {e}")
                    time.sleep(2)  # Allow final events to be rendered

                # Now stop card monitoring
                if self.card_monitor_thread:
                    self.card_monitor_thread._stop_event = True

                if self.ollama_manager and self.attributes["debug"]:
                    print(
                        f"[@ollama] process statuses: {self.ollama_manager.processes}"
                    )
                    print(
                        f"[@ollama] process runtime stats: {self.ollama_manager.stats}"
                    )
                    print(
                        f"[@ollama] Circuit Breaker status: {self.ollama_manager.circuit_breaker.get_status()}"
                    )

        return ollama_wrapper



import os
import json

__mf_promote_submodules__ = ["plugins.optuna"]


def auth():
    from metaflow.metaflow_config_funcs import init_config

    conf = init_config()
    if conf:
        headers = {"x-api-key": conf["METAFLOW_SERVICE_AUTH_KEY"]}
    else:
        headers = json.loads(os.environ["METAFLOW_SERVICE_HEADERS"])
    return headers


def get_deployment_db_access_endpoint(name: str):
    from ..apps.core.perimeters import PerimeterExtractor
    from ..apps.core.capsule import CapsuleApi

    perimeter, cap_url = PerimeterExtractor.during_metaflow_execution()
    deployment = CapsuleApi(cap_url, perimeter).get_by_name(name)
    if not deployment:
        raise Exception(f"No app deployment found with name `{name}`")

    if (
        "status" in deployment
        and "accessInfo" in deployment["status"]
        and "extraAccessUrls" in deployment["status"]["accessInfo"]
    ):
        for extra_url in deployment["status"]["accessInfo"]["extraAccessUrls"]:
            if extra_url["name"] == "in_cluster_db_access":
                db_url = extra_url["url"].replace("http://", "")
                return db_url

    raise Exception(f"No db access endpoint found for deployment `{name}`")


def get_db_url(app_name: str):
    """
    Example usage:
        >>> from metaflow.plugins.optuna import get_db_url
        >>> s = optuna.create_study(..., storage=get_db_url("optuna-dashboard"))
    """
    mf_token = auth()["x-api-key"]
    app_url = get_deployment_db_access_endpoint(app_name)
    return f"postgresql://userspace_default:{mf_token}@{app_url}/userspace_default?sslmode=disable"



from datetime import datetime
import json
import os
import pathlib
import requests
import sys
import time
from functools import partial
import shlex
from typing import Optional, List, Dict, Any, Tuple, Union
from .utils import TODOException, safe_requests_wrapper, MaximumRetriesExceeded
from .app_config import AppConfig, CAPSULE_DEBUG, AuthType
from . import experimental
from ._state_machine import (
    _capsule_worker_semantic_status,
    _capsule_worker_status_diff,
    CapsuleWorkerSemanticStatus,
    WorkerStatus,
    CapsuleStatus,
    DEPLOYMENT_READY_CONDITIONS,
    LogLine,
)


def _format_url_string(url):
    if url is None:
        return None

    if url.startswith("http://") or url.startswith("https://"):
        return url

    return f"https://{url}"


class CapsuleStateMachine:
    """
    - Every capsule create call will return a `identifier` and a `version` of the object.
    - Each update call will return a new version.
    - The status.currentlyServedVersion will be the version that is currently serving traffic.
    - The status.updateInProgress will be True if an upgrade is in progress.

    CapsuleState Transition:
    - Every capsule create call will return a `identifier` and a `version` of the object.
    - Happy Path:
        - First time Create :
            - wait for status.updateInProgress to be set to False
                - (interleaved) Poll the worker endpoints to check their status
                    - showcase how many workers are coming up if things are on the cli side.
                - If the user has set some flag like `--dont-wait-to-fully-finish` then we check the `status.currentlyServedVersion` to see if even one replica is ready to
                serve traffic.
            - once the status.updateInProgress is set to False, it means that the replicas are ready
        - Upgrade:
            - wait for status.updateInProgress to be set to False
                - (interleaved) Poll the worker endpoints to check their status and signal the user the number replicas coming up
                - If the user has set some flag like `--dont-wait-to-fully-finish` then we check the `status.currentlyServedVersion` to see if even one replica is ready to
                serve traffic.
    - Unhappy Path:
        - First time Create :
            - wait for status.updateInProgress to be set to False,
                - (interleaved) Poll the workers to check their status.
                    - If the worker pertaining the current deployment instance version is crashlooping then crash the deployment process with the error messages and logs.
        - Upgrade:
            - wait for status.updateInProgress to be set to False,
                - (interleaved) Poll the workers to check their status.
                    - If the worker pertaining the current deployment instance version is crashlooping then crash the deployment process with the error messages and logs.

    """

    def __init__(self, capsule_id: str, current_deployment_instance_version: str):
        self._capsule_id = capsule_id
        self._status_trail: List[Dict[str, Any]] = []
        self._current_deployment_instance_version = current_deployment_instance_version

    def get_status_trail(self):
        return self._status_trail

    def add_status(self, status: CapsuleStatus):
        assert type(status) == dict, "TODO: Make this check somewhere else"
        self._status_trail.append({"timestamp": time.time(), "status": status})

    @property
    def current_status(self):
        return self._status_trail[-1].get("status")

    @property
    def out_of_cluster_url(self):
        access_info = self.current_status.get("accessInfo", {}) or {}
        return _format_url_string(access_info.get("outOfClusterURL", None))

    @property
    def in_cluster_url(self):
        access_info = self.current_status.get("accessInfo", {}) or {}
        return _format_url_string(access_info.get("inClusterURL", None))

    @property
    def update_in_progress(self):
        return self.current_status.get("updateInProgress", False)

    @property
    def currently_served_version(self):
        return self.current_status.get("currentlyServedVersion", None)

    @property
    def ready_to_serve_traffic(self):
        if self.current_status.get("readyToServeTraffic", False):
            return any(
                i is not None for i in [self.out_of_cluster_url, self.in_cluster_url]
            )
        return False

    @property
    def available_replicas(self):
        return self.current_status.get("availableReplicas", 0)

    def report_current_status(self, logger):
        pass

    def save_debug_info(self, state_dir: str):
        debug_path = os.path.join(state_dir, f"debug_capsule_{self._capsule_id}.json")
        with open(debug_path, "w") as f:
            json.dump(self._status_trail, f, indent=4)


class CapsuleWorkersStateMachine:
    def __init__(
        self,
        capsule_id: str,
        end_state_capsule_version: str,
        deployment_mode: str = DEPLOYMENT_READY_CONDITIONS.ATLEAST_ONE_RUNNING,
        minimum_replicas: int = 1,
    ):
        self._capsule_id = capsule_id
        self._end_state_capsule_version = end_state_capsule_version
        self._deployment_mode = deployment_mode
        self._minimum_replicas = minimum_replicas
        self._status_trail: List[Dict[str, Union[float, List[WorkerStatus]]]] = []

    def get_status_trail(self):
        return self._status_trail

    def add_status(self, worker_list_response: List[WorkerStatus]):
        """
        worker_list_response: List[Dict[str, Any]]
            [
                {
                    "workerId": "c-4pqikm-659dd9ccdc-5hcwz",
                    "phase": "Running",
                    "activity": 0,
                    "activityDataAvailable": false,
                    "version": "0xhgaewiqb"
                },
                {
                    "workerId": "c-4pqikm-b8559688b-xk2jh",
                    "phase": "Pending",
                    "activity": 0,
                    "activityDataAvailable": false,
                    "version": "421h48qh95"
                }
            ]
        """
        self._status_trail.append(
            {"timestamp": time.time(), "status": worker_list_response}
        )

    def save_debug_info(self, state_dir: str):
        debug_path = os.path.join(
            state_dir, f"debug_capsule_workers_{self._capsule_id}_trail.json"
        )
        with open(debug_path, "w") as f:
            json.dump(self._status_trail, f, indent=4)

        status_path = os.path.join(
            state_dir, f"debug_capsule_workers_{self._capsule_id}_status.json"
        )
        with open(status_path, "w") as f:
            json.dump(self.current_version_deployment_status(), f, indent=4)

    def report_current_status(self, logger):
        if len(self._status_trail) == 0:
            return
        older_status = None
        if len(self._status_trail) >= 2:
            older_status = _capsule_worker_semantic_status(
                self._status_trail[-2].get("status"),
                self._end_state_capsule_version,
                self._minimum_replicas,
            )
        current_status = self.current_version_deployment_status()
        changes = _capsule_worker_status_diff(current_status, older_status)
        if len(changes) > 0:
            logger(*changes)

    @property
    def current_status(self) -> List[WorkerStatus]:
        return self._status_trail[-1].get("status")  # type: ignore

    def current_version_deployment_status(self) -> CapsuleWorkerSemanticStatus:
        return _capsule_worker_semantic_status(
            self.current_status, self._end_state_capsule_version, self._minimum_replicas
        )

    @property
    def is_crashlooping(self) -> bool:
        status = self.current_version_deployment_status()
        return status["status"]["at_least_one_crashlooping"]


class CapsuleInput:
    @classmethod
    def construct_exec_command(cls, commands: List[str]):
        commands = ["set -eEuo pipefail"] + commands
        command_string = "\n".join(commands)
        # First construct a base64 encoded string of the quoted command
        # One of the reasons we don't directly pass the command string to the backend with a `\n` join
        # is because the backend controller doesn't play nice when the command can be a multi-line string.
        # So we encode it to a base64 string and then decode it back to a command string at runtime to provide to
        # `bash -c`. The ideal thing to have done is to run "bash -c {shlex.quote(command_string)}" and call it a day
        # but the backend controller yields the following error:
        # `error parsing template: error converting YAML to JSON: yaml: line 111: mapping values are not allowed in this context`
        # So we go to great length to ensure the command is provided in base64 to avoid any issues with the backend controller.
        import base64

        encoded_command = base64.b64encode(command_string.encode()).decode()
        decode_cmd = f"echo {encoded_command} | base64 -d > ./_ob_app_run.sh"
        return (
            f"bash -c '{decode_cmd} && cat ./_ob_app_run.sh && bash ./_ob_app_run.sh'"
        )

    @classmethod
    def _marshal_environment_variables(cls, app_config: AppConfig):
        envs = app_config.get_state("environment", {}).copy()
        _return = []
        for k, v in envs.items():
            _v = v
            if isinstance(v, dict):
                _v = json.dumps(v)
            elif isinstance(v, list):
                _v = json.dumps(v)
            else:
                _v = str(v)
            _return.append(
                {
                    "name": k,
                    "value": _v,
                }
            )
        return _return

    @classmethod
    def from_app_config(cls, app_config: AppConfig):
        ## Replica settings
        replicas = app_config.get_state("replicas", {})
        fixed, _min, _max = (
            replicas.get("fixed"),
            replicas.get("min"),
            replicas.get("max"),
        )
        rpm = replicas.get("scaling_policy", {}).get("rpm", None)
        autoscaling_config = {}
        if rpm:
            autoscaling_config = {
                "requestRateBasedAutoscalingConfig": {"targetRequestsPerMinute": rpm}
            }
        if fixed is not None:
            _min, _max = fixed, fixed
        gpu_resource = app_config.get_state("resources").get("gpu")
        resources = {}
        shared_memory = app_config.get_state("resources").get("shared_memory")
        if gpu_resource:
            resources["gpu"] = gpu_resource
        if shared_memory:
            resources["sharedMemory"] = shared_memory

        _scheduling_config = {}
        if app_config.get_state("compute_pools", None):
            _scheduling_config["schedulingConfig"] = {
                "computePools": [
                    {"name": x} for x in app_config.get_state("compute_pools")
                ]
            }
        _description = app_config.get_state("description")
        _app_type = app_config.get_state("app_type")
        _final_info = {}
        if _description:
            _final_info["description"] = _description
        if _app_type:
            _final_info["endpointType"] = _app_type
        return {
            "perimeter": app_config.get_state("perimeter"),
            **_final_info,
            "codePackagePath": app_config.get_state("code_package_url"),
            "image": app_config.get_state("image"),
            "resourceIntegrations": [
                {"name": x} for x in app_config.get_state("secrets", [])
            ],
            "resourceConfig": {
                "cpu": str(app_config.get_state("resources").get("cpu")),
                "memory": str(app_config.get_state("resources").get("memory")),
                "ephemeralStorage": str(app_config.get_state("resources").get("disk")),
                **resources,
            },
            "autoscalingConfig": {
                "minReplicas": _min,
                "maxReplicas": _max,
                **autoscaling_config,
            },
            **_scheduling_config,
            "containerStartupConfig": {
                "entrypoint": cls.construct_exec_command(
                    app_config.get_state("commands")
                )
            },
            "environmentVariables": cls._marshal_environment_variables(app_config),
            # "assets": [{"name": "startup-script.sh"}],
            "authConfig": {
                "authType": app_config.get_state("auth").get("type"),
                "publicToDeployment": app_config.get_state("auth").get("public"),
            },
            "tags": [
                dict(key=k, value=v)
                for tag in app_config.get_state("tags", [])
                for k, v in tag.items()
            ],
            "port": app_config.get_state("port"),
            "displayName": app_config.get_state("name"),
            "forceUpdate": app_config.get_state("force_upgrade", False),
        }


class CapsuleApiException(Exception):
    def __init__(
        self,
        url: str,
        method: str,
        status_code: int,
        text: str,
        message: Optional[str] = None,
    ):
        self.url = url
        self.method = method
        self.status_code = status_code
        self.text = text
        self.message = message

    def __str__(self):
        return (
            f"CapsuleApiException: {self.url} [{self.method}]: Status Code: {self.status_code} \n\n {self.text}"
            + (f"\n\n {self.message}" if self.message else "")
        )


class CapsuleDeploymentException(Exception):
    def __init__(
        self,
        capsule_id: str,
        message: str,
    ):
        self.capsule_id = capsule_id
        self.message = message

    def __str__(self):
        return f"CapsuleDeploymentException: [{self.capsule_id}] :: {self.message}"


class CapsuleApi:
    def __init__(self, base_url: str, perimeter: str, logger_fn=None):
        self._base_url = self._create_base_url(base_url, perimeter)
        from metaflow.metaflow_config import SERVICE_HEADERS

        self._logger_fn = logger_fn
        self._request_headers = {
            **{"Content-Type": "application/json", "Connection": "keep-alive"},
            **(SERVICE_HEADERS or {}),
        }

    @staticmethod
    def _create_base_url(base_url: str, perimeter: str):
        return os.path.join(
            base_url,
            "v1",
            "perimeters",
            perimeter,
            "capsules",
        )

    def _wrapped_api_caller(self, method_func, *args, **kwargs):
        try:
            response = safe_requests_wrapper(
                method_func,
                *args,
                headers=self._request_headers,
                logger_fn=self._logger_fn,
                **kwargs,
            )
        except MaximumRetriesExceeded as e:
            raise CapsuleApiException(
                e.url,
                e.method,
                e.status_code,
                e.text,
                message=f"Maximum retries exceeded for {e.url} [{e.method}]",
            )
        if response.status_code >= 400:
            raise CapsuleApiException(
                args[0],
                method_func.__name__,
                response.status_code,
                response.text,
            )
        return response

    def create(self, capsule_input: dict):
        _data = json.dumps(capsule_input)
        response = self._wrapped_api_caller(
            requests.post,
            self._base_url,
            data=_data,
        )
        try:
            return response.json()
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                self._base_url,
                "post",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )

    def get(self, capsule_id: str):
        _url = os.path.join(self._base_url, capsule_id)
        response = self._wrapped_api_caller(
            requests.get,
            _url,
            retryable_status_codes=[409, 404],  # todo : verify me
            conn_error_retries=3,
        )
        try:
            return response.json()
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                _url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )

    def get_by_name(self, name: str, most_recent_only: bool = True):
        _url = os.path.join(self._base_url, f"?displayName={name}")
        response = self._wrapped_api_caller(
            requests.get,
            _url,
            retryable_status_codes=[409],  # todo : verify me
            conn_error_retries=3,
        )
        try:
            if most_recent_only:
                result = response.json()
                candidates = result["capsules"]
                if not candidates:
                    return None
                return sorted(
                    candidates, key=lambda x: x["metadata"]["createdAt"], reverse=True
                )[0]
            else:
                return response.json()
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                _url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )

    def list(self):
        response = self._wrapped_api_caller(
            requests.get,
            self._base_url,
            retryable_status_codes=[409],  # todo : verify me
            conn_error_retries=3,
        )
        try:
            response_json = response.json()
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                self._base_url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )
        if "capsules" not in response_json:
            raise CapsuleApiException(
                self._base_url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )
        return response_json.get("capsules", []) or []

    def delete(self, capsule_id: str):
        _url = os.path.join(self._base_url, capsule_id)
        response = self._wrapped_api_caller(
            requests.delete,
            _url,
            retryable_status_codes=[409],  # todo : verify me
        )
        if response.status_code >= 400:
            raise CapsuleApiException(
                _url,
                "delete",
                response.status_code,
                response.text,
            )

        if response.status_code == 200:
            return True
        return False

    def get_workers(self, capsule_id: str) -> List[Dict[str, Any]]:
        _url = os.path.join(self._base_url, capsule_id, "workers")
        response = self._wrapped_api_caller(
            requests.get,
            _url,
            retryable_status_codes=[409, 404],  # todo : verify me
            # Adding 404s because sometimes we might even end up getting 404s if
            # the backend cache is not updated yet. So on consistent 404s we should
            # just crash out.
            conn_error_retries=3,
        )
        try:
            return response.json().get("workers", []) or []
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                _url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )

    def logs(
        self, capsule_id: str, worker_id: str, previous: bool = False
    ) -> List[LogLine]:
        _url = os.path.join(self._base_url, capsule_id, "workers", worker_id, "logs")
        options = None
        if previous:
            options = {"previous": True}
        response = self._wrapped_api_caller(
            requests.get,
            _url,
            retryable_status_codes=[409],  # todo : verify me
            params=options,
        )
        try:
            return response.json().get("logs", []) or []
        except json.JSONDecodeError as e:
            raise CapsuleApiException(
                _url,
                "get",
                response.status_code,
                response.text,
                message="Capsule JSON decode failed",
            )

    def patch(self, capsule_id: str, patch_input: dict):
        capsule_response = self.get(capsule_id)
        if "spec" not in capsule_response or len(capsule_response.get("spec", {})) == 0:
            raise CapsuleApiException(
                self._base_url,
                "patch",
                403,
                "Capsule response of incorrect format",
            )

        spec = capsule_response.get("spec")
        spec.update(patch_input)
        return self.create(spec)


def list_and_filter_capsules(
    capsule_api: CapsuleApi, project, branch, name, tags, auth_type, capsule_id
):
    capsules = capsule_api.list()

    def _tags_match(tags, key, value):
        for t in tags:
            if t["key"] == key and t["value"] == value:
                return True
        return False

    def _all_tags_match(tags, tags_to_match):
        return all([_tags_match(tags, t["key"], t["value"]) for t in tags_to_match])

    def _filter_capsules(capsules, project, branch, name, tags, auth_type, capsule_id):
        _filtered_capsules = []
        for capsule in capsules:
            set_tags = capsule.get("spec", {}).get("tags", [])
            display_name = capsule.get("spec", {}).get("displayName", None)
            set_id = capsule.get("id", None)
            set_auth_type = (
                capsule.get("spec", {}).get("authConfig", {}).get("authType", None)
            )

            if auth_type and set_auth_type != auth_type:
                continue
            if project and not _tags_match(set_tags, "project", project):
                continue
            if branch and not _tags_match(set_tags, "branch", branch):
                continue
            if name and display_name != name:
                continue
            if tags and not _all_tags_match(set_tags, tags):
                continue
            if capsule_id and set_id != capsule_id:
                continue

            _filtered_capsules.append(capsule)
        return _filtered_capsules

    return _filter_capsules(
        capsules, project, branch, name, tags, auth_type, capsule_id
    )


from collections import namedtuple

CapsuleInfo = namedtuple("CapsuleInfo", ["info", "workers"])


class CapsuleDeployer:

    status: CapsuleStateMachine

    identifier = None

    # TODO: Current default timeout is very large of 5 minutes. Ideally we should have finished the deployed in less than 1 minutes.
    def __init__(
        self,
        app_config: AppConfig,
        base_url: str,
        create_timeout: int = 60 * 5,
        debug_dir: Optional[str] = None,
        success_terminal_state_condition: str = DEPLOYMENT_READY_CONDITIONS.ATLEAST_ONE_RUNNING,
        readiness_wait_time: int = 20,
        logger_fn=None,
    ):
        self._app_config = app_config
        self._capsule_api = CapsuleApi(
            base_url,
            app_config.get_state("perimeter"),
            logger_fn=logger_fn or partial(print, file=sys.stderr),
        )
        self._create_timeout = create_timeout
        self._logger_fn = logger_fn
        self._debug_dir = debug_dir
        self._capsule_deploy_response = None
        self._success_terminal_state_condition = success_terminal_state_condition
        self._readiness_wait_time = readiness_wait_time

    @property
    def url(self):
        return _format_url_string(
            ({} or self._capsule_deploy_response).get("outOfClusterUrl", None)
        )

    @property
    def capsule_api(self):
        return self._capsule_api

    @property
    def capsule_type(self):
        auth_type = self._app_config.get_state("auth", {}).get("type", AuthType.default)
        if auth_type == AuthType.BROWSER:
            return "App"
        elif auth_type == AuthType.API:
            return "Endpoint"
        else:
            raise TODOException(f"Unknown auth type: {auth_type}")

    @property
    def name(self):
        return self._app_config.get_state("name")

    def create_input(self):
        return experimental.capsule_input_overrides(
            self._app_config, CapsuleInput.from_app_config(self._app_config)
        )

    @property
    def current_deployment_instance_version(self):
        """
        The backend `create` call returns a version of the object that will be
        """
        if self._capsule_deploy_response is None:
            return None
        return self._capsule_deploy_response.get("version", None)

    def create(self):
        capsule_response = self._capsule_api.create(self.create_input())
        self.identifier = capsule_response.get("id")
        self._capsule_deploy_response = capsule_response
        return self.identifier

    def get(self):
        return self._capsule_api.get(self.identifier)

    def get_workers(self):
        return self._capsule_api.get_workers(self.identifier)

    def _backend_version_mismatch_check(
        self, capsule_response: dict, current_deployment_instance_version: str
    ):
        """
        - `capsule_response.version` contains the version of the object present in the database
        - `current_deployment_instance_version` contains the version of the object that was deployed by this instance of the deployer.
        In the situation that the versions of the objects become a mismatch then it means that current deployment process is not giving the user the
        output that they desire.
        """
        if capsule_response.get("version", None) != current_deployment_instance_version:
            raise CapsuleDeploymentException(
                self.identifier,  # type: ignore
                f"A capsule upgrade was triggered outside current deployment instance. Current deployment version was discarded. Current deployment version: {current_deployment_instance_version} and new version: {capsule_response.get('version', None)}",
            )

    def _monitor_worker_readiness(
        self,
        workers_sm: "CapsuleWorkersStateMachine",
    ):
        """returns True if the worker is crashlooping, False otherwise"""
        logger = self._logger_fn or partial(print, file=sys.stderr)
        for i in range(self._readiness_wait_time):
            time.sleep(1)
            workers_response = self.get_workers()
            workers_sm.add_status(workers_response)
            workers_sm.report_current_status(logger)
            if workers_sm.is_crashlooping:
                return True
        return False

    def _extract_logs_from_crashlooping_worker(
        self, workers_sm: "CapsuleWorkersStateMachine"
    ):
        def _extract_worker_id_of_crashlooping_worker(
            workers_status: List[WorkerStatus],
        ):
            for worker in workers_status:
                if worker["phase"] == "CrashLoopBackOff" or worker["phase"] == "Failed":
                    return worker["workerId"]
            return None

        worker_id = _extract_worker_id_of_crashlooping_worker(workers_sm.current_status)
        if worker_id is None:
            return None, None
        logs = self.capsule_api.logs(self.identifier, worker_id, previous=True)
        return logs, worker_id

    def _get_min_replicas(self):
        replicas = self._app_config.get_state("replicas", {})
        fixed, _min, _ = replicas.get("fixed"), replicas.get("min"), replicas.get("max")
        if fixed is not None:
            return fixed
        return _min

    def wait_for_terminal_state(
        self,
    ):
        """ """
        logger = self._logger_fn or partial(print, file=sys.stderr)
        state_machine = CapsuleStateMachine(
            self.identifier, self.current_deployment_instance_version
        )
        # min_replicas will always be present
        min_replicas = self._get_min_replicas()
        workers_state_machine = CapsuleWorkersStateMachine(
            self.identifier,
            self.current_deployment_instance_version,
            deployment_mode=self._success_terminal_state_condition,
            minimum_replicas=min_replicas,
        )
        self.status = state_machine
        for i in range(self._create_timeout):
            time.sleep(1)
            capsule_response = self.get()
            workers_response = self.get_workers()

            # We first need to check if someone has not upgraded the capsule under the hood and
            # the current deployment instance is invalid.
            self._backend_version_mismatch_check(
                capsule_response, self.current_deployment_instance_version  # type: ignore
            )
            state_machine.add_status(capsule_response.get("status", {}))  # type: ignore
            workers_state_machine.add_status(workers_response)
            state_machine.report_current_status(logger)

            workers_state_machine.report_current_status(logger)
            # Deployment readiness checks will determine what is the terminal state
            # of the workerstate machine. If we detect a terminal state in the workers,
            # then even if the capsule upgrade is still in progress we will end up crashing
            # the deployment.
            (
                capsule_ready,
                further_check_worker_readiness,
            ) = DEPLOYMENT_READY_CONDITIONS.check_readiness_condition(
                state_machine.current_status,
                workers_state_machine.current_version_deployment_status(),
                self._success_terminal_state_condition,
            )

            failure_condition_satisfied = (
                DEPLOYMENT_READY_CONDITIONS.check_failure_condition(
                    state_machine.current_status,
                    workers_state_machine.current_version_deployment_status(),
                )
            )
            if capsule_ready or failure_condition_satisfied:
                logger(
                    "💊 %s deployment status: %s "
                    % (
                        self.capsule_type.title(),
                        (
                            "in progress"
                            if state_machine.update_in_progress
                            else "completed"
                        ),
                    )
                )
                _further_readiness_check_failed = False
                if further_check_worker_readiness:
                    # HACK : monitor the workers for N seconds to make sure they are healthy
                    # this is a hack. Ideally we should implement a healthcheck as a first class citizen
                    # but it will take some time to do that so in the meanwhile a timeout set on the cli
                    # side will be really helpful.
                    logger(
                        "💊 Running last minute readiness check for %s..."
                        % self.identifier
                    )
                    _further_readiness_check_failed = self._monitor_worker_readiness(
                        workers_state_machine
                    )

                if CAPSULE_DEBUG:
                    logger(
                        f"[debug] 💊 {self.capsule_type} {self.identifier}: further_check_worker_readiness {_further_readiness_check_failed} | failure_condition_satisfied {failure_condition_satisfied}"
                    )

                # We should still check for failure state and crash if we detect something in the readiness check
                if failure_condition_satisfied or _further_readiness_check_failed:
                    # hit the logs endpoint for the worker and get the logs
                    # Print those logs out on the terminal
                    # raise an exception that should be caught gracefully by the cli
                    logs, worker_id = self._extract_logs_from_crashlooping_worker(
                        workers_state_machine
                    )
                    if logs is not None:
                        # todo: It would be really odd if the logs are not present and we discover something is crashlooping.
                        # Handle that condition later
                        logger(
                            *(
                                [
                                    f"💥 Worker ID ({worker_id}) is crashlooping. Please check the following logs for more information: "
                                ]
                                + ["\t" + l["message"] for l in logs]
                            )
                        )
                        raise CapsuleDeploymentException(
                            self.identifier,
                            f"Worker ID ({worker_id}) is crashlooping. Please check the logs for more information.",
                        )

                if state_machine.ready_to_serve_traffic:
                    logger(
                        "💊 %s %s is ready to serve traffic on the URL: %s"
                        % (
                            self.capsule_type,
                            self.identifier,
                            state_machine.out_of_cluster_url,
                        ),
                    )

                break

            if CAPSULE_DEBUG and self._debug_dir:
                state_machine.save_debug_info(self._debug_dir)
                workers_state_machine.save_debug_info(self._debug_dir)
                if i % 3 == 0:  # Every 3 seconds report the status
                    logger(
                        f"[debug] 💊 {self.capsule_type} {self.identifier} deployment status: {state_machine.current_status} | worker states: {workers_state_machine.current_status} | capsule_ready : {capsule_ready} | further_check_worker_readiness {further_check_worker_readiness}"
                    )

        # We will only check ready_to_serve_traffic under the following conditions:
        # If the readiness condition is not Async and min_replicas in this deployment
        # instance is < 0
        _is_async_readiness = (
            self._success_terminal_state_condition == DEPLOYMENT_READY_CONDITIONS.ASYNC
        )
        if (
            min_replicas > 0
            and not _is_async_readiness
            and not self.status.ready_to_serve_traffic
        ):
            raise CapsuleDeploymentException(
                self.identifier,
                f"Capsule {self.identifier} failed to be ready to serve traffic",
            )

        if CAPSULE_DEBUG and self._debug_dir:
            state_machine.save_debug_info(self._debug_dir)
            workers_state_machine.save_debug_info(self._debug_dir)
            logger(
                f"[debug] 💊 {self.capsule_type} {self.identifier} deployment status [on return]: {state_machine.current_status} | worker states: {workers_state_machine.current_status}"
            )

        return dict(
            id=self.identifier,
            auth_type=self.capsule_type,
            public_url=self.url,
            available_replicas=self.status.available_replicas,
            name=self.name,
            deployed_version=self.current_deployment_instance_version,
            deployed_at=datetime.now().isoformat(),
        )



"""
Typed Init Generator for ConfigMeta Classes

This module provides a mechanism to dynamically generate explicit typed classes
from ConfigMeta classes that IDEs can understand and provide autocomplete for.
"""

from typing import Any, Dict, List, Optional, Union, Type, Set

from .config_utils import ConfigMeta

import os

current_dir = os.path.dirname(__file__)

TYPED_DICT_IMPORT = """
import sys
from typing import TYPE_CHECKING

# on 3.8+ use the stdlib TypedDict;
# in TYPE_CHECKING blocks mypy/pyright still pick it up on older Pythons
if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    if TYPE_CHECKING:
        # for the benefit of type-checkers
        from typing import TypedDict  # noqa: F401
    # runtime no-op TypedDict shim
    class _TypedDictMeta(type):
        def __new__(cls, name, bases, namespace, total=True):
            # ignore total at runtime
            return super().__new__(cls, name, bases, namespace)

    class TypedDict(dict, metaclass=_TypedDictMeta):
        # Runtime stand-in for typing.TypedDict on <3.8.
        pass
"""


def collect_nested_configs_recursive(
    config_class: Type, visited: Optional[Set[str]] = None
) -> Dict[str, Type]:
    """
    Recursively collect all nested ConfigMeta classes from a config class.

    Args:
        config_class: A class that inherits from ConfigMeta
        visited: Set of already visited class names to avoid infinite recursion

    Returns:
        Dictionary mapping class names to ConfigMeta classes
    """
    if visited is None:
        visited = set()

    nested_configs = {}

    # Avoid infinite recursion by tracking visited classes
    if config_class.__name__ in visited:
        return nested_configs

    visited.add(config_class.__name__)

    # First pass: collect immediate nested configs
    for field_name, field_info in config_class._fields.items():
        if ConfigMeta.is_instance(field_info.field_type):
            nested_class = field_info.field_type
            nested_configs[nested_class.__name__] = nested_class

            # Recursively collect nested configs from this nested class
            deeper_nested = collect_nested_configs_recursive(
                nested_class, visited.copy()
            )
            nested_configs.update(deeper_nested)

    return nested_configs


def generate_typed_class_code(config_class: Type) -> str:
    """
    Generate the actual Python code for a typed class that IDEs can understand.

    Args:
        config_class: A class that inherits from ConfigMeta

    Returns:
        Python code string for the typed class
    """
    if not hasattr(config_class, "_fields"):
        raise ValueError(f"Class {config_class.__name__} is not a ConfigMeta class")

    class_name = f"Typed{config_class.__name__}"

    # Generate TypedDict for nested configs - now recursive
    nested_typeddict_code = []

    # Recursively collect all nested configs
    nested_configs = collect_nested_configs_recursive(config_class)

    # Generate TypedDict classes for all nested configs
    for nested_name, nested_class in nested_configs.items():
        dict_name = f"{nested_name}Dict"
        fields = []

        for field_name, field_info in nested_class._fields.items():
            field_type = _get_type_string(field_info.field_type, quote_config_meta=True)
            if not field_info.required:
                field_type = f"Optional[{field_type}]"
            fields.append(f"    {field_name}: {field_type}")

        typeddict_code = f"""class {dict_name}(TypedDict, total=False):
{chr(10).join(fields)}"""
        nested_typeddict_code.append(typeddict_code)

    # Generate __init__ method signature
    required_params = []
    optional_params = []
    all_assignments = []

    for field_name, field_info in config_class._fields.items():
        field_type = field_info.field_type

        # Handle nested ConfigMeta classes
        if ConfigMeta.is_instance(field_type):
            type_hint = f"Optional[{field_type.__name__}Dict]"
            param_line = f"        {field_name}: {type_hint} = None"
            optional_params.append(param_line)
        else:
            # All params will be set as options here even if the are required in the
            # configMeta
            type_hint = _get_type_string(field_type)
            param_line = f"        {field_name}: Optional[{type_hint}] = None"
            optional_params.append(param_line)

        all_assignments.append(f'            "{field_name}": {field_name}')

    # Combine required params first, then optional params
    all_params = required_params + optional_params

    # Generate the class code
    newline = "\n"
    comma_newline = ",\n"

    # Add **kwargs to the parameter list
    if all_params:
        params_with_kwargs = all_params + ["        **kwargs"]
    else:
        params_with_kwargs = ["        **kwargs"]

    class_code = f"""class {class_name}:
    def __init__(
        self,
{comma_newline.join(params_with_kwargs)}
    ) -> None:
        self._kwargs = {{
{comma_newline.join(all_assignments)}
        }}
        # Add any additional kwargs
        self._kwargs.update(kwargs)
        # Remove None values
        self._kwargs = {{k: v for k, v in self._kwargs.items() if v is not None}}
        self._config_class = {config_class.__name__}
        self._config = self.create_config()

    def create_config(self) -> {config_class.__name__}:
        return {config_class.__name__}.from_dict(self._kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return self._config.to_dict()"""

    # Combine all code
    full_code = []
    if nested_typeddict_code:
        full_code.extend(nested_typeddict_code)
        full_code.append("")  # Empty line
    full_code.append(class_code)

    return (newline + newline).join(full_code)


def _get_type_string(field_type: Type, quote_config_meta: bool = False) -> str:
    """Convert a type to its string representation for code generation.

    Args:
        field_type: The type to convert
        quote_config_meta: Whether to quote ConfigMeta type references for forward declarations
    """
    if field_type == str:
        return "str"
    elif field_type == int:
        return "int"
    elif field_type == float:
        return "float"
    elif field_type == bool:
        return "bool"
    elif ConfigMeta.is_instance(field_type):
        # Handle ConfigMeta classes by referencing their Dict type
        dict_type = f"{field_type.__name__}Dict"
        return f'"{dict_type}"' if quote_config_meta else dict_type
    elif hasattr(field_type, "__origin__"):
        # Handle generic types like List[str], Dict[str, str], etc.
        origin = field_type.__origin__
        args = getattr(field_type, "__args__", ())

        if origin == list:
            if args:
                return f"List[{_get_type_string(args[0], quote_config_meta)}]"
            return "List[Any]"
        elif origin == dict:
            if len(args) == 2:
                return f"Dict[{_get_type_string(args[0], quote_config_meta)}, {_get_type_string(args[1], quote_config_meta)}]"
            return "Dict[str, Any]"
        elif origin == Union:
            # Handle Optional types
            if len(args) == 2 and type(None) in args:
                non_none_type = args[0] if args[1] is type(None) else args[1]
                return f"Optional[{_get_type_string(non_none_type, quote_config_meta)}]"
            return f"Union[{', '.join(_get_type_string(arg, quote_config_meta) for arg in args)}]"

    # Default case - use the type name
    return getattr(field_type, "__name__", str(field_type))


def generate_typed_classes_module(
    config_classes: List[Type], module_name: str = "typed_configs"
) -> str:
    """
    Generate a complete Python module with typed classes for multiple ConfigMeta classes.

    Args:
        config_classes: List of ConfigMeta classes
        module_name: Name for the generated module

    Returns:
        Complete Python module code
    """
    imports = [
        "from typing import Optional, List, Dict, Any",
        "from .unified_config import "
        + ", ".join(cls.__name__ for cls in config_classes),
        TYPED_DICT_IMPORT,
    ]

    class_codes = []
    for config_class in config_classes:
        class_codes.append(generate_typed_class_code(config_class))

    # Use string concatenation instead of f-string with backslashes
    newline = "\n"
    module_code = (
        '"""'
        + newline
        + "Auto-generated typed classes for ConfigMeta classes."
        + newline
        + newline
        + "This module provides IDE-friendly typed interfaces for all configuration classes."
        + newline
        + "The reason we auto-generate this file is because we want to provide a bridge between what is the ConfigMeta classes and the typed programmatic interface."
        + newline
        + "The CoreConfig class is setup in a way that if any additionally params are missed out from being auto-generated "
        + "then it will not affect the core functionality of the programmatic API."
        + newline
        + "The new parameters will just not show up in IDE autocompletions."
        + newline
        + "It is fine if this file is not regularly updated by running the script in the .pre-commit-config.app-changes.yaml"
        + newline
        + "but it is recommended that this file not be deleted or manually edited."
        + newline
        + newline
        + '"""'
        + newline
        + newline
        + newline.join(imports)
        + newline
        + newline
        + (newline + newline).join(class_codes)
        + newline
    )

    return module_code


def create_typed_init_class_dynamic(config_class: Type) -> Type:
    """
    Dynamically create a typed init class with proper IDE support.

    This creates the class at runtime but with proper type annotations
    that IDEs can understand.
    """
    if not hasattr(config_class, "_fields"):
        raise ValueError(f"Class {config_class.__name__} is not a ConfigMeta class")

    class_name = f"Typed{config_class.__name__}"

    # Create the init method with proper signature
    def create_init_method():
        # Build the signature dynamically
        sig_params = []
        annotations: Dict[str, Any] = {"return": type(None)}

        for field_name, field_info in config_class._fields.items():
            field_type = field_info.field_type

            # Handle nested ConfigMeta classes
            if ConfigMeta.is_instance(field_type):
                field_type = Dict[str, Any]  # Use Dict for nested configs

            # Handle Optional fields
            if not field_info.required:
                field_type = Optional[field_type]

            annotations[field_name] = field_type

        def __init__(self, **kwargs):
            # Validate kwargs
            required_fields = {
                name for name, info in config_class._fields.items() if info.required
            }
            provided_fields = set(kwargs.keys())
            valid_fields = set(config_class._fields.keys())

            # Check required fields
            missing_fields = required_fields - provided_fields
            if missing_fields:
                raise ValueError(
                    f"Missing required fields: {', '.join(missing_fields)}"
                )

            # Check for unknown fields - but allow them for flexibility
            unknown_fields = provided_fields - valid_fields
            if unknown_fields:
                print(
                    f"Warning: Unknown fields will be passed through: {', '.join(unknown_fields)}"
                )

            self._kwargs = kwargs
            self._config_class = config_class

        # Set annotations
        __init__.__annotations__ = annotations
        return __init__

    def create_config(self):
        """Create and return the ConfigMeta class instance."""
        return config_class.from_dict(self._kwargs)

    def to_dict(self) -> Dict[str, Any]:
        """Return the raw kwargs as a dictionary."""
        return self._kwargs.copy()

    def __repr__(self) -> str:
        return f"{class_name}({self._kwargs})"

    # Create the class
    init_method = create_init_method()

    TypedClass = type(
        class_name,
        (object,),
        {
            "__init__": init_method,
            "create_config": create_config,
            "to_dict": to_dict,
            "__repr__": __repr__,
            "__module__": __name__,
            "__qualname__": class_name,
        },
    )

    return TypedClass


# Auto-generate and write typed classes to a file
def generate_typed_classes_file(output_file: Optional[str] = None):
    """
    Generate typed classes and write them to a file for IDE support.

    Args:
        output_file: Path to write the generated classes. If None, prints to stdout.
    """
    from .unified_config import CoreConfig

    config_classes = [CoreConfig]

    module_code = generate_typed_classes_module(config_classes)

    if output_file:
        with open(output_file, "w") as f:
            f.write(module_code)
        print(f"Generated typed classes written to {output_file}")
    else:
        print(module_code)


# Example usage and testing
if __name__ == "__main__":
    # Generate typed classes file
    generate_typed_classes_file(os.path.join(current_dir, "typed_configs.py"))



import requests
from requests.adapters import HTTPAdapter, Retry

BASE_URL = "https://api.ngc.nvidia.com/v2/orgs/zhxkmsaasxhw/"
POLL_SEC = 1


def _session(api_key):
    s = requests.Session()
    s.headers.update(
        {
            "Authorization": f"Bearer {api_key}",
            "Content-Type": "application/json",
        }
    )
    retry = Retry(total=5, backoff_factor=1.5, status_forcelist=[502, 503, 504])
    s.mount("https://", HTTPAdapter(max_retries=retry))
    return s


def _url(path):
    return BASE_URL.rstrip("/") + path


class NVCTClient:
    def __init__(self, api_key):
        self.sess = _session(api_key)

    # TODO: Handle https://outerboundsco.slack.com/archives/C05QGNR4E06/p1745970955540289
    def create(self, spec):
        r = self.sess.post(_url("/nvct/tasks"), json=spec, timeout=30)
        r.raise_for_status()
        return r.json().get("task", {}).get("id")

    def get(self, task_id):
        r = self.sess.get(_url(f"/nvct/tasks/{task_id}"), timeout=30)
        r.raise_for_status()
        return r.json().get("task", {})

    def cancel(self, task_id):
        r = self.sess.post(_url(f"/nvct/tasks/{task_id}/cancel"), timeout=30)
        r.raise_for_status()


class NVCTRequest(object):
    def __init__(self, name):
        self._spec = {}
        self._spec["name"] = name
        self._spec["gpuSpecification"] = {}
        self._spec["resultHandlingStrategy"] = "NONE"
        self._spec["terminationGracePeriodDuration"] = "PT10M"

    def container_image(self, image):
        self._spec["containerImage"] = image
        return self

    def container_args(self, args):
        self._spec["containerArgs"] = args
        return self

    def env(self, key, value):
        env_list = self._spec.setdefault("containerEnvironment", [])
        env_list.append({"key": key, "value": value})
        return self

    def gpu(self, gpu, instance_type, backend):
        gpu_spec = self._spec["gpuSpecification"]
        gpu_spec["gpu"] = gpu
        gpu_spec["instanceType"] = instance_type
        gpu_spec["backend"] = backend
        return self

    def max_runtime(self, iso_duration):
        self._spec["maxRuntimeDuration"] = iso_duration
        return self

    def max_queued(self, iso_duration="PT72H"):
        self._spec["maxQueuedDuration"] = iso_duration
        return self

    def termination_grace(self, iso_duration="PT10M"):
        self._spec["terminationGracePeriodDuration"] = iso_duration
        return self

    def extra(self, key, value):
        self._spec[key] = value
        return self

    def to_dict(self):
        return self._spec


class NVCTTask:
    def __init__(self, client: NVCTClient, spec):
        self.client = client
        self.spec = spec
        self.id = None
        self.record = None

    def submit(self):
        self.id = self.client.create(self.spec)
        return self.id

    def cancel(self):
        if not self.has_finished:
            self.client.cancel(self.id)

    @property
    def status(self):
        self.record = self.client.get(self.id)
        return self.record["status"]

    @property
    def is_waiting(self):
        return self.status == "QUEUED"

    @property
    def is_running(self):
        return self.status in {"RUNNING", "LAUNCHED"}

    @property
    def has_failed(self):
        return self.status in {"ERRORED", "CANCELED"}

    @property
    def has_succeeded(self):
        return self.status == "COMPLETED"

    @property
    def has_finished(self):
        return self.has_succeeded or self.has_failed



from .s3_proxy_decorator import (
    S3ProxyDecorator,
    NebiusS3ProxyDecorator,
    CoreWeaveS3ProxyDecorator,
)

__all__ = ["S3ProxyDecorator", "NebiusS3ProxyDecorator", "CoreWeaveS3ProxyDecorator"]



import os
import re
import time
import math
import shlex
import atexit

from metaflow import util
from metaflow.mflog import (
    BASH_SAVE_LOGS,
    bash_capture_logs,
    export_mflog_env_vars,
    tail_logs,
    get_log_tailer,
)

from .nvct import NVCTClient, NVCTTask, NVCTRequest
from .exceptions import (
    NvctKilledException,
    NvctExecutionException,
    NvctTaskFailedException,
)

# Constants for Metaflow logs
LOGS_DIR = "$PWD/.logs"
STDOUT_FILE = "mflog_stdout"
STDERR_FILE = "mflog_stderr"
STDOUT_PATH = os.path.join(LOGS_DIR, STDOUT_FILE)
STDERR_PATH = os.path.join(LOGS_DIR, STDERR_FILE)
NVCT_WRAPPER = "/usr/local/bin/nvct-wrapper.sh"


class NvctRunner:
    def __init__(
        self,
        metadata,
        datastore,
        environment,
        gpu_type,
        instance_type,
        backend,
        ngc_api_key,
    ):
        self.metadata = metadata
        self.datastore = datastore
        self.environment = environment
        self.gpu_type = gpu_type
        self.instance_type = instance_type
        self.backend = backend
        self._ngc_api_key = ngc_api_key
        self.client = None
        self.task = None
        atexit.register(lambda: self.task.cancel() if hasattr(self, "task") else None)

    def launch_task(
        self,
        step_name,
        step_cli,
        task_spec,
        code_package_sha,
        code_package_url,
        code_package_ds,
        env={},
        max_runtime="PT7H",  # <8H allowed for GFN backend
        max_queued="PT120H",  # 5 days
    ):
        mflog_expr = export_mflog_env_vars(
            datastore_type=code_package_ds,
            stdout_path=STDOUT_PATH,
            stderr_path=STDERR_PATH,
            **task_spec,
        )
        init_cmds = self.environment.get_package_commands(
            code_package_url, code_package_ds
        )
        init_expr = " && ".join(init_cmds)
        step_expr = bash_capture_logs(
            " && ".join(
                self.environment.bootstrap_commands(step_name, code_package_ds)
                + [step_cli]
            )
        )
        cmd_str = "mkdir -p %s && %s && %s && %s; c=$?; %s; exit $c" % (
            LOGS_DIR,
            mflog_expr,
            init_expr,
            step_expr,
            BASH_SAVE_LOGS,
        )

        # Add optional initialization script execution
        cmd_str = (
            '${METAFLOW_INIT_SCRIPT:+eval \\"${METAFLOW_INIT_SCRIPT}\\"} && %s'
            % cmd_str
        )

        cmd_str = shlex.split('bash -c "%s"' % cmd_str)[-1]

        def modify_python_c(match):
            content = match.group(1)
            # Escape double quotes within the python -c command
            content = content.replace('"', r"\"")
            # Replace outermost double quotes with single quotes
            return 'python -c "%s"' % content

        # Convert python -c single quotes to double quotes
        cmd_str = re.sub(r"python -c '(.*?)'", modify_python_c, cmd_str)
        cmd_str = cmd_str.replace("'", '"')
        # Create the final command with outer single quotes to pass to NVCT wrapper
        nvct_cmd = f"{NVCT_WRAPPER} bash -c '{cmd_str}'"

        flow_name = task_spec.get("flow_name")
        run_id = task_spec.get("run_id")
        task_id = task_spec.get("task_id")
        retry_count = task_spec.get("retry_count")
        task_name = f"{flow_name}-{run_id}-{step_name}-{task_id}-{retry_count}"

        if self.backend != "GFN":
            # if maxRuntimeDuration exceeds 8 hours for a Task on the GFN backend,
            # the request will be rejected.
            # (https://docs.nvidia.com/cloud-functions/user-guide/latest/cloud-function/tasks.html#create-task)
            ## thus, if it is non GFN backend, we increase it to 3 days
            max_runtime = "PT72H"

        request = (
            NVCTRequest(task_name)
            .container_image("nvcr.io/zhxkmsaasxhw/nvct-base:2.0-jovyan")
            .container_args(nvct_cmd)
            .gpu(
                gpu=self.gpu_type,
                instance_type=self.instance_type,
                backend=self.backend,
            )
            .max_runtime(max_runtime)
            .max_queued(max_queued)
        )

        for k, v in env.items():
            if v is not None:
                request.env(k, str(v))

        self.client = NVCTClient(self._ngc_api_key)
        self.task = NVCTTask(self.client, request.to_dict())

        self.task.submit()
        return self.task.id

    def wait_for_completion(self, stdout_location, stderr_location, echo=None):
        if not self.task:
            raise NvctExecutionException("No task has been launched")

        def update_delay(secs_since_start):
            # this sigmoid function reaches
            # - 0.1 after 11 minutes
            # - 0.5 after 15 minutes
            # - 1.0 after 23 minutes
            # in other words, the user will see very frequent updates
            # during the first 10 minutes
            sigmoid = 1.0 / (1.0 + math.exp(-0.01 * secs_since_start + 9.0))
            return 0.5 + sigmoid * 30.0

        def wait_for_launch(task):
            status = task.status
            echo(
                "Task is starting (%s)..." % status,
                "stderr",
                _id=task.id,
            )

            t = time.time()
            start_time = time.time()
            while task.is_waiting:
                new_status = task.status
                if status != new_status or (time.time() - t) > 30:
                    status = new_status
                    echo(
                        "Task is starting (%s)..." % status,
                        "stderr",
                        _id=task.id,
                    )
                    t = time.time()
                time.sleep(update_delay(time.time() - start_time))

        _make_prefix = lambda: b"[%s] " % util.to_bytes(self.task.id)
        stdout_tail = get_log_tailer(stdout_location, self.datastore.TYPE)
        stderr_tail = get_log_tailer(stderr_location, self.datastore.TYPE)

        # 1) Loop until the job has started
        wait_for_launch(self.task)

        echo(
            "Task is starting (%s)..." % self.task.status,
            "stderr",
            _id=self.task.id,
        )

        # 2) Tail logs until the job has finished
        tail_logs(
            prefix=_make_prefix(),
            stdout_tail=stdout_tail,
            stderr_tail=stderr_tail,
            echo=echo,
            has_log_updates=lambda: self.task.is_running,
        )

        if self.task.has_failed:
            raise NvctTaskFailedException(
                f"Task failed with status: {self.task.status}. This could be a transient error. Use @retry to retry."
            )
        else:
            if self.task.is_running:
                # Kill the job if it is still running by throwing an exception.
                raise NvctKilledException("Task failed!")
            echo(
                f"Task finished with status: {self.task.status}",
                "stderr",
                _id=self.task.id,
            )
